# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit test stub for ontosPy

Run like this:

$ python -m ontospy.tests.test_shapes

"""

from __future__ import print_function

import time

import unittest, os, sys
from .. import *
from ..core import *
from ..core.utils import *


from .context import TEST_RDF_FOLDER



# sanity check
printDebug(f"-------------------\nOntospy {VERSION}\n-------------------")


class TestShapes(unittest.TestCase):


	printDebug(f"""\n=================\n
	\nTEST SHACL SHAPES: Loading selected ontologies in => {TEST_RDF_FOLDER} 
	\nFor each model shapes description is extracted and printed.
	\n\n=================""", bg="blue", fg="white")

	time.sleep(3)

	candidates = [
		TEST_RDF_FOLDER + "scigraph/" ,
		TEST_RDF_FOLDER + "uco_monolithic.ttl" ,
	]


	def test1_local_shapes(self):

		"""
		Check if the shapes in the SciGraph onto are loaded properly
		"""

		for f in self.candidates:

			printDebug("\n*****\nTest: loading resource... > %s\n*****" % str(f), bg="green")

			o = Ontospy(f, verbose=False)

			for el in o.stats():
				printDebug("%s : %d" % (el[0], el[1]))

			for s in o.all_shapes:
				printDebug("\nSHAPE: %s" % str(s), "green")
				if s.targetClasses:
					for x in s.targetClasses:
						printDebug(".....hasTargetClass: " + str(x))
						printDebug("     Reverse link: the class has %d associated shape." % len(x.all_shapes))
				else:
					printDebug("..... has no target class!")


			for c in o.all_classes:
				if c.shapedProperties:
					printDebug("Class: %s" % str(c), "green")
					for x in c.shapedProperties:
						print(".....hasProperty: " + str(x["property"]))
						print("     the property has %s associated shapes." % x["shape"])
				else:
					pass




if __name__ == "__main__":
	unittest.main()




from jinja2 import Environment, PackageLoader, select_autoescape
import re
from markupsafe import Markup, escape
from jinja2 import pass_eval_context
from ontospy.core.utils import slugify

#
# SET UP JINJA2
#
env = Environment(
    loader=PackageLoader('ontospy.gendocs.media', 'templates'),
    autoescape=select_autoescape(['html', 'xml']),
    extensions=['jinja2_time.TimeExtension']
)


#############################
####### JINJA2 CUSTOM FILTERS
###


#
# FILTER
#

def slugify_filter(value):
    """A filter for legacy django templates to call that returns a slugified value."""
    return slugify(value)  # from ontospy.core.utils

env.filters['slugify'] = slugify_filter


#
# FILTER
#

_paragraph_re = re.compile(r'(?:\r\n|\r|\n){2,}')

@pass_eval_context
def linebreaks_filter(eval_ctx, value):
    """A filter for legacy django templates."""
    try:
        if not value:
            return ""
        result = u'\n\n'.join(u'<p>%s</p>' % p.replace('\n', '<br>\n')
                        for p in _paragraph_re.split(escape(value)))
        if eval_ctx.autoescape:
            result = Markup(result)
        return result
    except:
        return value

env.filters['linebreaks'] = linebreaks_filter

#
# FILTER
#


def capfirst_filter(value):
    """A filter for legacy django templates to call that returns a capitalized value."""
    try:
        return value.capitalize()
    except:
        return value

env.filters['capfirst'] = capfirst_filter



#
# FILTER
#


def add_filter(value, integer_n):
    """A filter for legacy django templates."""
    try:
        return value + integer_n
    except:
        return value

env.filters['add'] = add_filter



#
# FILTER
#


def truncatewords_filter(data, l=20):
    """A filter for legacy django templates."""
    try:
        data_list = data.split()
        info = (" ".join(data_list[:l]) + '..') if len(data_list) > l else data
        return info
    except:
        return data


env.filters['truncatewords'] = truncatewords_filter




#
# FILTER
#


def d3_dendogram_height_filter(tot_objects):
    """A filter to generate dynamically the min height of a dendogram."""
    n = 50 * tot_objects
    if n < 800:
        return 800
    else:
        return n


env.filters['d3_dendogram_height'] = d3_dendogram_height_filter




# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Ontospy Shell Module
michele.pasin@gmail.com

# docs:
# https://docs.python.org/2/library/cmd.html
# https://hg.python.org/cpython/file/2.7/Lib/cmd.py
# http://pymotw.com/2/cmd/


"""


import sys
import os
import cmd
import random
import shutil
import platform
import urllib as urllib2
from pyfiglet import Figlet
from colorama import Fore, Style

from subprocess import PIPE, Popen
PY2 = sys.version < '3'
WINDOWS = os.name == 'nt'
EOL = '\r\n' if WINDOWS and not PY2 else '\n'


from .. import *  # load top level __init__
from ..VERSION import VERSION

from ..core import ONTOSPY_LOCAL
from ..core import manager
from ..core import actions
from ..core.ontospy import Ontospy
from ..core.utils import *


from .shell_quotes import *  # quotes


f = Figlet(font='slant')
_intro_ = """***
The Command Line Ontology Browser (%s)
***											  """

STARTUP_MESSAGE = f.renderText('Ontospy') + Style.BRIGHT + _intro_ % VERSION + Style.RESET_ALL


def _get_prompt(onto="", entity=""):
    """
    Global util that changes the prompt contextually
    :return: [Ontospy]>(cidoc_crm_v5.0...)>(class:E1.CRM_Entity)>
    """
    base_text, onto_text, entity_text = "", "", ""
    base_color, onto_color, entity_color = Fore.RED + Style.BRIGHT, Fore.BLACK + Style.DIM, Fore.BLACK

    if not onto and not entity:
        base_text = base_color + '[Ontospy]' + Style.RESET_ALL

    if onto and not entity:
        _tmp = onto_color + '(%s)' % onto
        onto_text = _tmp + Style.RESET_ALL

    if entity:
        _tmp = onto_color + Style.BRIGHT + '(%s)' % truncate(onto, 15)
        onto_text = _tmp + Style.RESET_ALL

        _tmp2 = entity_color + '(%s:%s)' % (entity['type'], entity['name'])
        entity_text = "-" + _tmp2 + Style.RESET_ALL

    return base_text + onto_text + entity_text + "> "


class Shell(cmd.Cmd):
    """Simple command processor example."""

    prompt = _get_prompt()
    intro = "Type 'help' to get started, TAB to explore commands.\n"

    doc_header = 'Commands available (type `help <command>` to get help):'
    misc_header = 'Miscellaneous'
    undoc_header = 'Undocumented commands'

    ruler = '-'
    maxcol = 80

    INFO_OPTS = ['namespaces', 'toplayer', 'parents',
                 'children', 'ancestors', 'descendants', 'inferred_usage']
    SERIALIZE_OPTS = ['xml', 'n3', 'turtle', 'nt', 'pretty-xml', 'json-ld']
    LS_OPTS = ['ontologies', 'classes', 'properties', 'concepts']
    TREE_OPTS = ['classes', 'properties', 'concepts']
    GET_OPTS = ['ontology', 'class', 'property', 'concept']
    FILE_OPTS = ['rename', 'delete']
    IMPORT_OPTS = ['uri', 'file',  'repo', 'starter-pack', ]
    VISUALIZE_OPTS = []

    def __init__(self, uri=None):
        """
        self.current = {'file' : filename, 'fullpath' : fullpath, 'graph': g}
        self.currentEntity = {'name' : obj.locale or obj.uri, 'object' : obj, 'type' : 'class'}
                                                                    # or 'property' or 'concept'
        """
        # init repo if necessary
        manager.get_or_create_home_repo()
        # useful vars
        self.LOCAL = ONTOSPY_LOCAL
        self.LOCAL_MODELS = manager.get_home_location()
        self.all_ontologies = manager.get_localontologies()
        self.current = None
        self.currentEntity = None
        if uri:
            self._load_ontology(uri, preview_mode=True)
        cmd.Cmd.__init__(self)

    # BASE CLASSE OVERRIDES:
    # --------

    def emptyline(self):
        """ override default behaviour of running last command """
        pass

    def print_topics(self, header, cmds, cmdlen, maxcol):
        """Override 'print_topics' so that you can exclude EOF and shell.
            2016-02-12: added to test, copied from
            https://github.com/xlcnd/isbntools/blob/master/isbntools/bin/repl.py
        """
        if header:
            if cmds:
                self.stdout.write("%s\n" % str(header))
                if self.ruler:
                    self.stdout.write("%s\n" % str(self.ruler * len(header)))
                self.columnize(cmds, maxcol - 1)
                self.stdout.write("\n")

    def default(self, line):
        "default message when a command is not recognized"
        foo = ["Don't recognize that command. Try 'help' for some suggestions.",
               "That looks like the wrong command", "Are you sure you mean that? Try 'help' for some suggestions."]
        self._print(random.choice(foo))

    # HELPER METHODS
    # --------

    def _print(self, ms, style="TIP"):
        """ abstraction for managing color printing """
        styles1 = {'IMPORTANT': Style.BRIGHT,
                   'TIP': Style.DIM,
                   'URI': Style.BRIGHT,
                   'TEXT': Fore.GREEN,
                   'MAGENTA': Fore.MAGENTA,
                   'BLUE': Fore.BLUE,
                   'GREEN': Fore.GREEN,
                   'RED': Fore.RED,
                   'DEFAULT': Style.DIM,
                   }
        try:
            printInfo(styles1[style] + ms + Style.RESET_ALL)
        except:
            printInfo(styles1['DEFAULT'] + ms + Style.RESET_ALL)

    def _printM(self, messages):
        """print a list of strings - for the mom used only by stats printout"""
        if len(messages) == 2:
            printInfo(Style.BRIGHT + messages[0] + Style.RESET_ALL +
                  Fore.BLUE + messages[1] + Style.RESET_ALL)
        else:
            printInfo("Not implemented")

    def _joinedQnames(self, _list):
        """util for returning a string joinin names of entities *used only in info command*"""
        try:
            s = "; ".join([p.qname for p in _list])
        except:
            s = "; ".join([p for p in _list])
        return s

    def _clear_screen(self):
        """ http://stackoverflow.com/questions/18937058/python-clear-screen-in-shell """
        if platform.system() == "Windows":
            tmp = os.system('cls')  # for window
        else:
            tmp = os.system('clear')  # for Linux
        return True

    def _printTriples(self, entity):
        """ display triples """
        self._print("----------------", "TIP")
        self._print(unicode(entity.uri), "IMPORTANT")
        for x in entity.triples:
            self._print("=> " + unicode(x[1]), "MAGENTA")
            self._print(".... " + unicode(x[2]), "GREEN")
        self._print("----------------", "TIP")

    def _print_entity_intro(self, g=None, entity=None, first_time=True):
        """after a selection, prints on screen basic info about onto or entity, plus change prompt
        2015-10-18: removed the sound
        2016-01-18: entity is the shell wrapper around the ontospy entity
        """
        if entity:
            self._clear_screen()
            obj = entity['object']
            self._print("Loaded %s: <%s>" % (entity['type'].capitalize(), str(obj.uri)), "TIP")
            self._print("----------------", "TIP")
            # self._print(obj.bestDescription(), "TEXT")
            if first_time:
                self.prompt = _get_prompt(self.current['file'], self.currentEntity)
        elif g:
            self._printDescription(False)
            if first_time:
                self.prompt = _get_prompt(self.current['file'])

    def _printStats(self, graph, hrlinetop=False):
        """ shotcut to pull out useful info for interactive use
        2016-05-11: note this is a local version of graph.printStats()
        """
        if hrlinetop:
            self._print("----------------", "TIP")
        self._print("Ontologies......: %d" % len(graph.all_ontologies), "TIP")
        self._print("Classes.........: %d" % len(graph.all_classes), "TIP")
        self._print("Properties......: %d" % len(graph.all_properties), "TIP")
        self._print("..annotation....: %d" % len(graph.all_properties_annotation), "TIP")
        self._print("..datatype......: %d" % len(graph.all_properties_datatype), "TIP")
        self._print("..object........: %d" % len(graph.all_properties_object), "TIP")
        self._print("Concepts(SKOS)..: %d" % len(graph.all_skos_concepts), "TIP")
        self._print("----------------", "TIP")

    def _printDescription(self, hrlinetop=True):
        """generic method to print out a description"""
        if hrlinetop:
            self._print("----------------")
        NOTFOUND = "[not found]"
        if self.currentEntity:
            obj = self.currentEntity['object']
            label = obj.bestLabel() or NOTFOUND
            description = obj.bestDescription() or NOTFOUND
            printInfo(Style.BRIGHT + "OBJECT TYPE: " + Style.RESET_ALL +
                  Fore.BLACK + uri2niceString(obj.rdftype) + Style.RESET_ALL)
            printInfo(Style.BRIGHT + "URI        : " + Style.RESET_ALL +
                  Fore.GREEN + "<" + unicode(obj.uri) + ">" + Style.RESET_ALL)
            printInfo(Style.BRIGHT + "TITLE      : " + Style.RESET_ALL +
                  Fore.BLACK + label + Style.RESET_ALL)
            printInfo(Style.BRIGHT + "DESCRIPTION: " + Style.RESET_ALL +
                  Fore.BLACK + description + Style.RESET_ALL)

        else:
            self._clear_screen()
            self._print("Graph: <" + self.current['fullpath'] + ">", 'TIP')
            self._print("----------------", "TIP")
            self._printStats(self.current['graph'])
            for obj in self.current['graph'].all_ontologies:
                printInfo(Style.BRIGHT + "Ontology URI: " + Style.RESET_ALL +
                      Fore.RED + "<%s>" % str(obj.uri) + Style.RESET_ALL)
                # self._print("==> Ontology URI: <%s>" % str(obj.uri), "IMPORTANT")
                # self._print("----------------", "TIP")
                label = obj.bestLabel() or NOTFOUND
                description = obj.bestDescription() or NOTFOUND
                printInfo(Style.BRIGHT + "Title       : " + Style.RESET_ALL +
                      Fore.BLACK + label + Style.RESET_ALL)
                printInfo(Style.BRIGHT + "Description : " + Style.RESET_ALL +
                      Fore.BLACK + description + Style.RESET_ALL)
        self._print("----------------", "TIP")
        # self._print("----------------", "TIP")

    def _printTaxonomy(self, hrlinetop=True):
        """
        print a local taxonomy for the object
        """
        if not self.currentEntity:  # ==> ontology level
            return
        if hrlinetop:
            self._print("----------------")
        self._print("TAXONOMY:", "IMPORTANT")
        x = self.currentEntity['object']
        parents = x.parents()

        if not parents:
            if self.currentEntity['type'] == 'class':
                self._print("owl:Thing")
            elif self.currentEntity['type'] == 'property':
                self._print("RDF:Property")
            elif self.currentEntity['type'] == 'concept':
                self._print("SKOS:Concept")
            else:
                pass
        else:
            for p in parents:
                self._print(p.qname)
        self._print("..." + x.qname, "TEXT")
        for c in x.children():
            self._print("......" + c.qname)
        self._print("----------------")

    def _printClassDomain(self, hrlinetop=True, print_inferred=False):
        """
        print more informative stats about the object
        2016-06-14: added inferred option
        """
        if not self.currentEntity:  # ==> ontology level
            return
        x = self.currentEntity['object']
        if self.currentEntity['type'] == 'class':
            if hrlinetop:
                self._print("----------------")
            self._print("DOMAIN OF:", "IMPORTANT")
            self._print("[%d] Explicitly declared" % len(x.domain_of), "IMPORTANT")
            for i in x.domain_of:
                if i.ranges:
                    ranges = ",".join([y.qname if hasattr(y, "qname") else str(y)
                                       for y in i.ranges])
                else:
                    ranges = "owl:Thing"
                    # print( Style.RESET_ALL + " => " + Fore.MAGENTA +
                printInfo(Fore.GREEN + x.qname + Style.RESET_ALL + " => " + Fore.MAGENTA +
                      i.qname + Style.RESET_ALL + " => " + Style.DIM + ranges +
                      Style.RESET_ALL)

            # inferred stuff
            if print_inferred:
                for _dict in x.domain_of_inferred:
                    _class = list(_dict.items())[0][0]
                    _propslist = list(_dict.items())[0][1]
                    if _class.id != x.id and len(_propslist):  # print only inferred properties
                        self._print("[%d] Inherited from [%s]" % (len(_propslist), _class.qname),
                                    "IMPORTANT")
                        for i in _propslist:
                            if i.ranges:
                                ranges = ",".join([y.qname if hasattr(y, "qname")
                                                   else str(y) for y in i.ranges])
                            else:
                                ranges = "owl:Thing"
                                # print(Style.RESET_ALL + " => " + Fore.MAGENTA +
                            printInfo(Fore.GREEN + x.qname + Style.RESET_ALL + " => " + Fore.MAGENTA +
                                  i.qname + Style.RESET_ALL + " => " + Style.DIM +
                                  ranges + Style.RESET_ALL)

            self._print("----------------")
        return

    def _printClassRange(self, hrlinetop=True, print_inferred=False):
        """
        print(more informative stats about the object)
        2016-06-14: added inferred option
        """
        if not self.currentEntity:  # ==> ontology level
            return
        x = self.currentEntity['object']
        if self.currentEntity['type'] == 'class':
            if hrlinetop:
                self._print("----------------")
            self._print("RANGE OF:", "IMPORTANT")
            self._print("[%d] Explicitly declared" % len(x.range_of), "IMPORTANT")
            for i in x.range_of:
                if i.domains:
                    domains = ",".join([y.qname if hasattr(y, "qname") else str(y)
                                        for y in i.domains])
                else:
                    domains = "owl:Thing"
                printInfo(Style.DIM + domains + Style.RESET_ALL + " => " + Fore.MAGENTA + i.qname +
                      Style.RESET_ALL + " => " + Fore.GREEN + x.qname + Style.RESET_ALL)

            # inferred stuff
            if print_inferred:
                for _dict in x.range_of_inferred:
                    _class = list(_dict.items())[0][0]
                    _propslist = list(_dict.items())[0][1]
                    if _class.id != x.id and len(_propslist):  # print only inferred properties
                        self._print("[%d] Inherited from [%s]" % (len(_propslist), _class.qname),
                                    "IMPORTANT")
                        for i in _propslist:
                            if i.domains:
                                domains = ",".join(
                                    [y.qname if hasattr(y, "qname") else str(y) for y in i.domains])
                            else:
                                domains = "owl:Thing"
                            printInfo(Style.DIM + domains + Style.RESET_ALL + " => " + Fore.MAGENTA +
                                  i.qname + Style.RESET_ALL + " => " + Fore.GREEN + x.qname + Style.RESET_ALL)

            self._print("----------------")
        return

    def _printPropertyDomainRange(self, hrlinetop=True):
        """
        print(more informative stats about the object)
        """
        if not self.currentEntity:  # ==> ontology level
            return
        x = self.currentEntity['object']
        if self.currentEntity['type'] == 'property':
            if hrlinetop:
                self._print("----------------")
            self._print("USAGE:", "IMPORTANT")
            domains = x.domains
            ranges = x.ranges
            if x.domains:
                for d in x.domains:
                    # for domain/ranges which are not declared classes
                    if hasattr(d, "qname"):
                        _name = d.qname
                    else:
                        _name = str(d)
                    self._print(_name)
            else:
                self._print("owl:Thing")
            self._print("  " + x.qname, "TEXT")
            if x.ranges:
                for d in x.ranges:
                    # for domain/ranges which are not declared classes
                    if hasattr(d, "qname"):
                        _name = d.qname
                    else:
                        _name = str(d)
                    self._print("  " + "   => " + _name)
            else:
                self._print("  " + "   => " + "owl:Thing")
            self._print("----------------")
        return

    def _printInstances(self, hrlinetop=True):
        """
        print(more informative stats about the object)
        """
        if not self.currentEntity:  # ==> ontology level
            return
        x = self.currentEntity['object']
        if self.currentEntity['type'] == 'class':
            if hrlinetop:
                self._print("----------------")
            self._print("INSTANCES: [%d]" % len(x.instances, "IMPORTANT"))
            for i in x.instances:
                self._print(i.qname)
            self._print("----------------")
        return

    def _printSourceCode(self, hrlinetop=True):
        """
        print(more informative stats about the object)
        """
        if not self.currentEntity:  # ==> ontology level
            return
        x = self.currentEntity['object']
        if hrlinetop:
            self._print("----------------")

        self._print("Source:", "IMPORTANT")
        self.do_serialize("turtle")
        self._print("----------------")

        return

    def _selectFromList(self, _list, using_pattern=True, objtype=None):
        """
        Generic method that lets users pick an item from a list via input
        *using_pattern* flag to know if we're showing all choices or not
        Note: the list items need to be Ontospy entities.
        <objtype>: if specified, it allows incremental search by keeping specifying a
        different pattern.
        """
        if not _list:
            self._print("No matching items.", "TIP")
            return None
        if using_pattern and len(_list) == 1:  # removed
            pass
            # return _list[0]
        if using_pattern:
            self._print("%d matching items: \n--------------" % len(_list), "TIP")
        else:
            self._print("%d items available: \n--------------" % len(_list), "TIP")
        counter = 1
        _temp = []
        for el in _list:
            if hasattr(el, 'qname'):
                _temp += [Fore.BLUE + Style.BRIGHT + "[%d] " %
                          counter + Style.RESET_ALL + str(el.qname)]
            elif hasattr(el, 'uri'):
                _temp += [Fore.BLUE + Style.BRIGHT + "[%d] " %
                          counter + Style.RESET_ALL + str(el.uri)]
            else:
                _temp += [Fore.BLUE + Style.BRIGHT + "[%d] " % counter + Style.RESET_ALL + str(el)]
            counter += 1
        pprint2columns(_temp)

        self._print("--------------")
        self._print("Please select one option by entering its number, or a keyword to filter: ")
        var = input()
        if var == "":
            return None
        elif var.isdigit():
            try:
                var = int(var)
                return _list[var-1]
            except:
                self._print("Selection not valid")
                return None
        elif objtype:
            # continuos patter matching on list (only for certain object types)
            if objtype == 'ontology':
                self._select_ontology(var)
            elif objtype == 'class':
                self._select_class(var)
            elif objtype == 'property':
                self._select_property(var)
            elif objtype == 'concept':
                self._select_concept(var)
            return

    def _next_ontology(self):
        """Dynamically retrieves the next ontology in the list"""
        currentfile = self.current['file']
        try:
            idx = self.all_ontologies.index(currentfile)
            return self.all_ontologies[idx+1]
        except:
            return self.all_ontologies[0]

    # MAIN METHODS
    # --------

    def _load_ontology(self, filename, preview_mode=False):
        """
        Loads an ontology

        Unless preview_mode=True, it is always loaded from the local repository
        note: if the ontology does not have a cached version, it is created

        preview_mode: used to pass a URI/path to be inspected without saving it locally
        """
        if not preview_mode:
            fullpath = self.LOCAL_MODELS + filename
            g = manager.get_pickled_ontology(filename)
            if not g:
                g = manager.do_pickle_ontology(filename)
        else:
            fullpath = filename
            filename = os.path.basename(os.path.normpath(fullpath))
            g = Ontospy(fullpath, verbose=True)
        self.current = {'file': filename, 'fullpath': fullpath, 'graph': g}
        self.currentEntity = None
        self._print_entity_intro(g)

    def _select_ontology(self, line):
        """try to select an ontology NP: the actual load from FS is in <_load_ontology> """
        try:
            var = int(line)  # it's a string
            if var in range(1, len(self.all_ontologies)+1):
                self._load_ontology(self.all_ontologies[var-1])
        except ValueError:
            out = []
            for each in self.all_ontologies:
                if line in each:
                    out += [each]
            choice = self._selectFromList(out, line, "ontology")
            if choice:
                self._load_ontology(choice)

    def _select_class(self, line):
        """
        try to match a class and load it from the graph
        NOTE: the g.get_class(pattern) method does the heavy lifting
        """
        g = self.current['graph']
        if not line:
            out = g.all_classes
            using_pattern = False
        else:
            using_pattern = True
            if line.isdigit():
                line = int(line)
            out = g.get_class(line)
        if out:
            if type(out) == type([]):
                choice = self._selectFromList(out, using_pattern, "class")
                if choice:
                    self.currentEntity = {'name': choice.locale or choice.uri,
                                          'object': choice, 'type': 'class'}
            else:
                self.currentEntity = {'name': out.locale or out.uri, 'object': out, 'type': 'class'}
            # ..finally:
            if self.currentEntity:
                self._print_entity_intro(entity=self.currentEntity)

        else:
            printInfo("not found")

    def _select_property(self, line):
        """try to match a property and load it"""
        g = self.current['graph']
        if not line:
            out = g.all_properties
            using_pattern = False
        else:
            using_pattern = True
            if line.isdigit():
                line = int(line)
            out = g.get_property(line)
        if out:
            if type(out) == type([]):
                choice = self._selectFromList(out, using_pattern, "property")
                if choice:
                    self.currentEntity = {'name': choice.locale or choice.uri,
                                          'object': choice, 'type': 'property'}

            else:
                self.currentEntity = {'name': out.locale or out.uri,
                                      'object': out, 'type': 'property'}

            # ..finally:
            if self.currentEntity:
                self._print_entity_intro(entity=self.currentEntity)
        else:
            printInfo("not found")

    def _select_concept(self, line):
        """try to match a class and load it"""
        g = self.current['graph']
        if not line:
            out = g.all_skos_concepts
            using_pattern = False
        else:
            using_pattern = True
            if line.isdigit():
                line = int(line)
            out = g.get_skos(line)
        if out:
            if type(out) == type([]):
                choice = self._selectFromList(out, using_pattern, "concept")
                if choice:
                    self.currentEntity = {'name': choice.locale or choice.uri,
                                          'object': choice, 'type': 'concept'}
            else:
                self.currentEntity = {'name': out.locale or out.uri,
                                      'object': out, 'type': 'concept'}
            # ..finally:
            if self.currentEntity:
                self._print_entity_intro(entity=self.currentEntity)

        else:
            printInfo("not found")

    def _delete_file(self, line=""):
        """	Delete an ontology
            2016-04-11: not a direct command anymore """

        if not self.all_ontologies:
            self._help_nofiles()

        else:
            out = []
            for each in self.all_ontologies:
                if line in each:
                    out += [each]
            choice = self._selectFromList(out, line)
            if choice:
                fullpath = self.LOCAL_MODELS + "/" + choice
                if os.path.isfile(fullpath):

                    self._print("--------------")
                    self._print("Are you sure? [Y/N]")
                    var = input()
                    if var == "y" or var == "Y":
                        os.remove(fullpath)
                        manager.del_pickled_ontology(choice)
                        self._print("<%s> was deleted succesfully." % choice)
                        self.all_ontologies = manager.get_localontologies()
                    else:
                        return

                else:
                    self._print("File not found.")
                # delete
                if self.current and self.current['fullpath'] == fullpath:
                    self.current = None
                    self.currentEntity = None
                    self.prompt = _get_prompt()

        return

    def _rename_file(self, line=""):
        """Rename an ontology
            2016-04-11: not a direct command anymore """

        if not self.all_ontologies:
            self._help_nofiles()
        else:
            out = []
            for each in self.all_ontologies:
                if line in each:
                    out += [each]
            choice = self._selectFromList(out, line)
            if choice:
                fullpath = self.LOCAL_MODELS + "/" + choice
                printDebug(fullpath)
                if os.path.isfile(fullpath):

                    self._print("--------------")
                    self._print("Please enter a new name for <%s>, including the extension (blank=abort)"
                                % choice)
                    var = input()
                    if var:
                        try:
                            os.rename(fullpath, self.LOCAL_MODELS + "/" + var)
                            manager.rename_pickled_ontology(choice, var)
                            self._print("<%s> was renamed succesfully." % choice)
                            self.all_ontologies = manager.get_localontologies()
                        except:
                            self._print("Not a valid name. An error occurred.")
                            return
                    else:
                        return

                else:
                    self._print("File not found.")
                # delete
                if self.current and self.current['fullpath'] == fullpath:
                    self.current = None
                    self.currentEntity = None
                    self.prompt = _get_prompt()

        return

    # COMMANDS
    # --------
    # NOTE: all commands should start with 'do_' and must pass 'line'

    def do_ls(self, line):
        """Shows entities of a given kind."""
        opts = self.LS_OPTS
        line = line.split()
        _pattern = ""

        if len(line) == 0:
            # default contextual behaviour [2016-03-01]
            if not self.current:
                line = ["ontologies"]
            elif self.currentEntity:
                self.do_info("")
                return
            else:
                line = ["classes"]

        if (not line) or (line[0] not in opts):
            self.help_ls()
            return
            # self._print("Usage: ls [%s]" % "|".join([x for x in opts]))

        elif line[0] == "ontologies":
            if not self.all_ontologies:
                self._help_nofiles()
            else:
                self._select_ontology(_pattern)

        elif line[0] in opts and not self.current:
            self._help_noontology()
            return

        elif line[0] == "classes":
            g = self.current['graph']

            if g.all_classes:
                self._select_class(_pattern)
            else:
                self._print("No classes available.")

        elif line[0] == "properties":
            g = self.current['graph']
            if g.all_properties:
                self._select_property(_pattern)
            else:
                self._print("No properties available.")

        elif line[0] == "concepts":
            g = self.current['graph']
            if g.all_skos_concepts:
                self._select_concept(_pattern)
            else:
                self._print("No concepts available.")

        else:  # should never arrive here
            pass

    def do_tree(self, line):
        """Shows entities of a given kind."""
        opts = self.TREE_OPTS
        line = line.split()
        _pattern = ""

        if not self.current:
            self._help_noontology()
            return

        if len(line) == 0:
            # default contextual behaviour [2016-03-01]
            line = ["classes"]

        if line[0] not in opts:
            self.help_tree()
            return

        elif line[0] == "classes":
            g = self.current['graph']
            if g.all_classes:
                g.printClassTree(showids=False, labels=False, showtype=True)
                self._print("----------------", "TIP")
            else:
                self._print("No classes available.")

        elif line[0] == "properties":
            g = self.current['graph']
            if g.all_properties:
                g.printPropertyTree(showids=False, labels=False, showtype=True)
            else:
                self._print("No properties available.")

        elif line[0] == "concepts":
            g = self.current['graph']
            if g.all_skos_concepts:
                g.printSkosTree(showids=False, labels=False, showtype=True)
            else:
                self._print("No concepts available.")

        else:  # should never arrive here
            pass

    def do_get(self, line):
        """Finds entities matching a given string pattern. \nOptions: [ ontologies | classes | properties | concepts ]"""
        line = line.split()
        _pattern = ""
        if len(line) > 1:
            _pattern = line[1]
        opts = self.GET_OPTS

        if (not line) or (line[0] not in opts) or (not _pattern):
            self.help_get()
            return
            # self._print("Usage: get [%s] <name>" % "|".join([x for x in opts]))

        elif line[0] == "ontology":
            if not self.all_ontologies:
                self._help_nofiles()
            else:
                self._select_ontology(_pattern)

        elif line[0] in opts and not self.current:
            self._help_noontology()
            return

        elif line[0] == "class":
            g = self.current['graph']
            if g.all_classes:
                self._select_class(_pattern)
            else:
                self._print("No classes available.")

        elif line[0] == "property":
            g = self.current['graph']
            if g.all_properties:
                self._select_property(_pattern)
            else:
                self._print("No properties available.")

        elif line[0] == "concept":
            g = self.current['graph']
            if g.all_skos_concepts:
                self._select_concept(_pattern)
            else:
                self._print("No concepts available.")

        else:  # should never arrive here
            pass

    def do_info(self, line):
        """Inspect the current entity and display a nice summary of key properties"""
        # opts = [ 'namespaces', 'description', 'overview', 'toplayer', 'parents', 'children', 'stats', 'triples' ]
        opts = self.INFO_OPTS

        if not self.current:
            self._help_noontology()
            return

        line = line.split()
        g = self.current['graph']

        # get arg, or default to 'overview'
        if not line:
            self._printDescription()
            self._printTaxonomy(False)
            self._printClassDomain(False)
            self._printClassRange(False)
            self._printPropertyDomainRange(False)
            # self._printSourceCode(False)
            if self.currentEntity and self.currentEntity['type'] == 'class':
                self._print("Tip: type 'info inferred_usage' to show inherited properties\n----------------")

        elif line[0] == "inferred_usage":
            if self.currentEntity:  # @todo only for classes?
                self._printDescription()
                self._printTaxonomy(False)
                self._printClassDomain(False, True)
                self._printClassRange(False, True)
                self._printPropertyDomainRange(False)
            # self._printSourceCode(False)
            return

        elif line[0] == "toplayer":
            if g.toplayer_classes:
                self._print("Top Classes\n----------------", "IMPORTANT")
            for x in g.toplayer_classes:
                self._print(x.qname)
            if g.toplayer_properties:
                self._print("\nTop Properties\n----------------", "IMPORTANT")
                for x in g.toplayer_properties:
                    self._print(x.qname)
            if g.toplayer_skos:
                self._print("\nTop Concepts (SKOS)\n----------------", "IMPORTANT")
                for x in g.toplayer_skos:
                    self._print(x.qname)

        elif line[0] == "namespaces":
            for x in self.current['graph'].namespaces:
                self._print("@prefix %s: <%s> ." % (x[0], x[1]))

        elif line[0] == "parents":
            if self.currentEntity and self.currentEntity['object'].parents():
                for x in self.currentEntity['object'].parents():
                    self._print(x.qname)
            else:
                self._print("No parents. This is a top level entity.")

        elif line[0] == "children":
            if self.currentEntity and self.currentEntity['object'].children():
                for x in self.currentEntity['object'].children():
                    self._print(x.qname)
            else:
                self._print("No children. This is a leaf node.")

        elif line[0] == "ancestors":
            if self.currentEntity and self.currentEntity['object'].ancestors():
                for x in self.currentEntity['object'].ancestors():
                    self._print(x.qname)
            else:
                self._print("No ancestors. This is a top level entity.")

        elif line[0] == "descendants":
            if self.currentEntity and self.currentEntity['object'].descendants():
                for x in self.currentEntity['object'].descendants():
                    self._print(x.qname)
            else:
                self._print("No descendants. This is a leaf node.")

            return

    def do_visualize(self, line):
        """Visualize an ontology - ie wrapper for export command"""

        if not self.current:
            self._help_noontology()
            return

        line = line.split()

        try:
            # from ..viz.actions import action_visualize
            from ..gendocs.actions import action_visualize
        except:
            self._print("This command requires the ontodocs package: `pip install ontodocs`")
            return

        import webbrowser
        url = action_visualize(args=self.current['file'], fromshell=True)
        if url:
            webbrowser.open(url)
        return

    def do_import(self, line):
        """Import an ontology"""

        line = line.split()

        if line and line[0] == "starter-pack":
            actions.action_bootstrap()

        elif line and line[0] == "uri":
            self._print(
                "------------------\nEnter a valid graph URI: (e.g. http://www.w3.org/2009/08/skos-reference/skos.rdf)")
            var = input()
            if var:
                if var.startswith("http"):
                    try:
                        actions.action_import(var)
                    except:
                        self._print(
                            "OPS... An Unknown Error Occurred - Aborting installation of <%s>" % var)
                else:
                    self._print("Not valid. TIP: URIs should start with 'http://'")

        elif line and line[0] == "file":
            self._print(
                "------------------\nEnter a full file path: (e.g. '/Users/mike/Desktop/journals.ttl')")
            var = input()
            if var:
                try:
                    actions.action_import(var)
                except:
                    self._print(
                        "OPS... An Unknown Error Occurred - Aborting installation of <%s>" % var)

        elif line and line[0] == "repo":
            actions.action_webimport()

        else:
            self.help_import()

        self.all_ontologies = manager.get_localontologies()
        return

    def do_file(self, line):
        """PErform some file operation"""
        opts = self.FILE_OPTS

        if not self.all_ontologies:
            self._help_nofiles()
            return

        line = line.split()

        if not line or line[0] not in opts:
            self.help_file()
            return

        if line[0] == "rename":
            self._rename_file()
        elif line[0] == "delete":
            self._delete_file()
        else:
            return

    def do_serialize(self, line):
        """Serialize an entity into an RDF flavour"""
        opts = self.SERIALIZE_OPTS

        if not self.current:
            self._help_noontology()
            return

        line = line.split()
        g = self.current['graph']

        if not line:
            line = ['turtle']

        if line[0] not in opts:
            self.help_serialize()
            return

        elif self.currentEntity:
            self.currentEntity['object'].printSerialize(line[0])

        else:
            self._print(g.rdf_source(format=line[0]))
            # 2016-05-27: was like this before
            # for o in g.all_ontologies:
            # 	o.printSerialize(line[0])

    def do_next(self, line):
        """Jump to the next entities (ontology, class or property) depending on context"""
        if not self.current:
            printInfo("Please select an ontology first. E.g. use the 'ls ontologies' or 'get ontology <name>' commands.")
        elif self.currentEntity:
            g = self.current['graph']
            if self.currentEntity['type'] == 'class':
                nextentity = g.nextClass(self.currentEntity['object'].uri)
                self._select_class(str(nextentity.uri))
            elif self.currentEntity['type'] == 'property':
                nextentity = g.nextProperty(self.currentEntity['object'].uri)
                self._select_property(str(nextentity.uri))
            elif self.currentEntity['type'] == 'concept':
                nextentity = g.nextConcept(self.currentEntity['object'].uri)
                self._select_concept(str(nextentity.uri))
            else:
                printInfo("Not implemented")
        else:
            if len(self.all_ontologies) > 1:
                nextonto = self._next_ontology()
                self._load_ontology(nextonto)
            else:
                self._print("Only one ontology available in repository.")

    def do_back(self, line):
        "Go back one step. From entity => ontology; from ontology => ontospy top level."
        if self.currentEntity:
            self.currentEntity = None
            self.prompt = _get_prompt(self.current['file'])
        else:
            self.current = None
            self.prompt = _get_prompt()

    def do_quit(self, line):
        "Quit: exit the Ontospy shell"
        self._clear_screen()
        return True

    def do_zen(self, line):
        """Inspiring quotes for the working ontologist"""
        _quote = random.choice(QUOTES)
        # print(_quote['source'])
        printInfo(Style.DIM + unicode(_quote['text']))
        printInfo(Style.BRIGHT + unicode(_quote['source']) + Style.RESET_ALL)

    # 2016-02-12: method taken from https://github.com/xlcnd/isbntools/blob/master/isbntools/bin/repl.py
    # 2016-04-25: hidden
    def _do_shell(self, line):
        """Send a command to the Unix shell.\n==> Usage: shell ls ~"""
        if not line:
            return
        sp = Popen(line,
                   shell=True,
                   stdin=PIPE,
                   stdout=PIPE,
                   stderr=PIPE,
                   close_fds=not WINDOWS)
        (fo, fe) = (sp.stdout, sp.stderr)
        if PY2:
            out = fo.read().strip(EOL)
            err = fe.read().strip(EOL)
        else:
            out = fo.read().decode("utf-8")
            err = fe.read().decode("utf-8")
        if out:
            printInfo(out)
            return
        if err:
            printInfo(err.replace('isbn_', ''))

    # HELP METHODS
    # --------

    def help_ls(self):
        txt = "List available graphs or entities .\n"
        txt += "==> Usage: ls [%s]" % "|".join([x for x in self.LS_OPTS])
        txt += "\n\nNote: ls is contextual. If you do not pass it any argument, it returns info based on the currently active object.\n"
        self._print(txt)

    def help_tree(self):
        txt = "Print the type hierarchy for an entity type.\n"
        txt += "==> Usage: tree [%s]" % "|".join([x for x in self.TREE_OPTS])
        # txt += "\n\nNote: tree is contextual. If you do not pass it any argument, it returns info based on the currently active object.\n"
        self._print(txt)

    def help_import(self):
        txt = "Import an ontology from a remote repository or directory.\n"
        txt += "==> Usage: import [%s]" % "|".join([x for x in self.IMPORT_OPTS])
        self._print(txt)

    def help_visualize(self):
        txt = "Visualize the currenlty selected ontology using an HTML template.\n"
        txt += "==> Usage: visualize "
        self._print(txt)

    def help_file(self):
        txt = "Perform some operations on the files in the local repository.\n"
        txt += "==> Usage: file [%s]" % "|".join([x for x in self.FILE_OPTS])
        self._print(txt)

    def help_serialize(self):
        txt = "Serialize an entity into an RDF flavour.\n"
        txt += "==> Usage: serialize [%s]" % "|".join([x for x in self.SERIALIZE_OPTS])
        self._print(txt)

    def help_get(self):
        txt = "Finds entities matching a given string pattern.\n"
        txt += "==> Usage: get [%s] <name>" % "|".join([x for x in self.GET_OPTS])
        self._print(txt)

    def help_info(self):
        txt = "Display information about an entity e.g. ontology, class etc..\n"
        txt += "==> Usage: info OR info [%s]" % "|".join([x for x in self.INFO_OPTS])
        self._print(txt)

    def _help_noontology(self):
        """starts with underscore so that it doesnt appear with help methods"""
        txt = "No graph selected. Please load a graph first.\n"
        txt += "==> E.g. use the 'ls ontologies' or 'get ontology <name>' commands."
        self._print(txt)

    def _help_nofiles(self):
        """starts with underscore so that it doesnt appear with help methods"""
        txt = "No files available in your local repository.\n"
        txt += "==> Use the 'import starter-pack' command to get started."
        self._print(txt)

    # AUTOCOMPLETE METHODS
    # --------

    def complete_ls(self, text, line, begidx, endidx):
        """completion for ls command"""

        options = self.LS_OPTS

        if not text:
            completions = options
        else:
            completions = [f
                           for f in options
                           if f.startswith(text)
                           ]
        return completions

    def complete_tree(self, text, line, begidx, endidx):
        """completion for ls command"""

        options = self.TREE_OPTS

        if not text:
            completions = options
        else:
            completions = [f
                           for f in options
                           if f.startswith(text)
                           ]
        return completions

    def complete_get(self, text, line, begidx, endidx):
        """completion for find command"""

        options = self.GET_OPTS

        if not text:
            completions = options
        else:
            completions = [f
                           for f in options
                           if f.startswith(text)
                           ]
        return completions

    def complete_info(self, text, line, begidx, endidx):
        """completion for info command"""

        opts = self.INFO_OPTS

        if not text:
            completions = opts
        else:
            completions = [f
                           for f in opts
                           if f.startswith(text)
                           ]
        return completions

    def complete_import(self, text, line, begidx, endidx):
        """completion for serialize command"""

        opts = self.IMPORT_OPTS

        if not text:
            completions = opts
        else:
            completions = [f
                           for f in opts
                           if f.startswith(text)
                           ]
        return completions

    def complete_serialize(self, text, line, begidx, endidx):
        """completion for serialize command"""

        opts = self.SERIALIZE_OPTS

        if not text:
            completions = opts
        else:
            completions = [f
                           for f in opts
                           if f.startswith(text)
                           ]
        return completions

    def complete_visualize(self, text, line, begidx, endidx):
        """completion for file command"""

        opts = self.VISUALIZE_OPTS

        if not text:
            completions = opts
        else:
            completions = [f
                           for f in opts
                           if f.startswith(text)
                           ]
        return completions

    def complete_file(self, text, line, begidx, endidx):
        """completion for file command"""

        opts = self.FILE_OPTS

        if not text:
            completions = opts
        else:
            completions = [f
                           for f in opts
                           if f.startswith(text)
                           ]
        return completions


def main():
    """ standalone line script """

    printInfo("Ontospy " + VERSION)

    Shell()._clear_screen()
    printInfo(Style.BRIGHT + "** Ontospy Interactive Ontology Browser " + VERSION + " **" + Style.RESET_ALL)
    # manager.get_or_create_home_repo()
    Shell().cmdloop()
    raise SystemExit(1)


if __name__ == '__main__':
    import sys
    try:
        main()
    except KeyboardInterrupt as e:  # Ctrl-C
        raise e



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit test stub for ontosPy

Run like this:

$ python -m ontospy.tests.test_methods

"""

from __future__ import print_function
import time

import unittest, os, sys
from .. import *
from ..core import *
from ..core.utils import *



from .context import TEST_RDF_FOLDER



# sanity check
printDebug(f"-------------------\nOntospy {VERSION}\n-------------------")




class SampleCustomEntity(ontospy.RdfEntity):

    def __init__(self, uri, rdftype=None, namespaces=None, ext_model=False, pref_title="qname", pref_lang="en"):
        super(SampleCustomEntity, self).__init__(uri, rdftype, namespaces, ext_model, pref_title, pref_lang)

    def __repr__(self):
        return "<SampleCustomEntity *%s*>" % ( self.uri)

    def disjointWith(self):
        """
        Example: pull out disjoint with statements
        """
        pred = "http://www.w3.org/2002/07/owl#disjointWith"
        return self.getValuesForProperty(pred)





class TestMethods(unittest.TestCase):


	printDebug(f"""\n=================\n
	\nTEST Methods: checking specific ontospy methods 
	\n\n=================""", bg="blue", fg="white")

	time.sleep(3)


	# load sample ontologies

	f = TEST_RDF_FOLDER + "pizza.ttl"
	printDebug("\n*****\n ..loading local ontology > %s\n*****" % str(f), "important")
	o = Ontospy(f, verbose=True, pref_title="label")
	
	f = TEST_RDF_FOLDER + "multilingual.ttl"
	printDebug("\n*****\n ..loading local ontology > %s\n*****" % str(f), "important")
	o2 = Ontospy(f, verbose=True, pref_title="qname", pref_lang="en")




	def test0(self):
		"""
		Class methods
		"""
		printDebug("\n=================\nTEST 0: Checking the <class> displays", bg="green")

		for c in self.o.all_classes:
			print("URI: ", c.uri)
			print("RDFTYPE: ", c.rdftype)
			print("BEST LABEL: ", c.bestLabel())
			print("TITLE: ", c.title)
			print("===")

		printDebug("Test completed succesfully.\n", "green")


	def test1(self):
		"""
		Instances method
		"""
		printDebug("\n=================\nTEST 1: Checking the <instances> method", bg="green")

		for c in self.o.all_classes:
			# c.describe()
			if c.instances:
				print("CLASS: " + c.uri + " " + c.title)
				print("INSTANCES: ")
				for el in c.instances:
					print(el.uri, el.qname)
					print(el.getValuesForProperty("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"))
		
		printDebug("Test completed succesfully.\n", "green")


	def test2(self):
		"""
		getValuesForProperty
		"""
		printDebug("\n=================\nTEST 2: Checking the <getValuesForProperty> method", bg="green")

		for c in self.o.all_classes[:3]:
			print("CLASS: ")
			print(c.uri, c.qname, c.title)
			print("RDF:TYPE VALUES: ")
			print(c.getValuesForProperty("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"))

		printDebug("Test completed succesfully.\n", "green")
	
	def test3(self):
		"""
		build_entity_from_uri
		"""
		printDebug("\n=================\nTEST 3: Checking the <build_entity_from_uri> method", bg="green")

		e = self.o.build_entity_from_uri("http://www.co-ode.org/ontologies/pizza/pizza.owl#Germany")
		print("URI: ", e)
		print("RDFTYPE: ", e.rdftype)
		print("BEST LABEL: ", e.bestLabel())
		print("TITLE: ", e.title)
		print("RDF SOURCE: ")
		print(e.rdf_source())
		printDebug("Test completed succesfully.\n", "green")


	def test4(self):
		"""
		build_entity_from_uri - SampleCustomEntity
		"""
		printDebug("\n=================\nTEST 4: Checking the <build_entity_from_uri> method using a SampleCustomEntity class ", bg="green")

		e = self.o.build_entity_from_uri("http://www.co-ode.org/ontologies/pizza/pizza.owl#FruttiDiMare", SampleCustomEntity)
		print("URI: ", e)
		print("RDFTYPE: ", e.rdftype)
		print("BEST LABEL: ", e.bestLabel())
		print("TITLE: ", e.title)
		print("OWL DISJOINT WITH: ")
		print("\n".join([x for x in e.disjointWith()]))
		printDebug("Test completed succesfully.\n", "green")



	def test5(self):
		"""
		Pref label and pref language parameters
		"""

		printDebug("\n=================\nTEST 5-1: pref_title=qname / pref_lang=en", bg="green")
		for c in self.o2.all_classes:
			print("URI: ", c.uri)
			print("RDFTYPE: ", c.rdftype)
			print("BEST LABEL: ", c.bestLabel())
			print("TITLE: ", c.title)
			print("===")

		printDebug("\n=================\nTEST 5-2: pref_title=label / pref_lang=it", bg="green")
		for c in self.o2.all_classes:
			print("URI: ", c.uri)
			print("RDFTYPE: ", c.rdftype)
			print("BEST LABEL: ", c.bestLabel())
			print("TITLE: ", c.title)
			print("===")

		printDebug("\n=================\nTEST 5-3: pref_title=label / pref_lang=es", bg="green")
		for c in self.o2.all_classes:
			print("URI: ", c.uri)
			print("RDFTYPE: ", c.rdftype)
			print("BEST LABEL: ", c.bestLabel())
			print("TITLE: ", c.title)
			print("===")

		printDebug("Test completed succesfully.\n", "green")

	
	print("Success.\n")









if __name__ == "__main__":
	unittest.main()



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit test stub for ontosPy

Test Quick: use this file to quickly run scripts/tests which will then be integrated into proper tests

"""

import click 

from .. import *
from ..core import *
from ..core.utils import *


from .context import TEST_RDF_FOLDER


# sanity check
print("-------------------\nOntospy ", VERSION, "\n-------------------")



@click.command()
@click.argument('test_number', nargs=1)
def main(test_number):

	test_number = int(test_number)


	if test_number == 1:

		print("=================\n*** QUICK TEST 1 ***\n=================\n")

		f = TEST_RDF_FOLDER + "paper.jsonld"

		o = Ontospy(f, verbose=True, rdf_format="json-ld", hide_implicit_types=False, hide_base_schemas=False, hide_implicit_preds=False)
		print(f)			



	if test_number == 2:

		print("=================\n*** QUICK TEST 2 ***\n=================\n")

		uri, title = "http://examples.com", "My ontology"
		printDebug(click.style("[%d]" % 1, fg='blue') +
				click.style(uri + " ==> ", fg='black') +
				click.style(title, fg='red'))


		from colorama import Fore, Style

		printDebug(Fore.BLUE + Style.BRIGHT + "[%d]" % 1 + 
              Style.RESET_ALL + uri + " ==> " + Fore.RED + title + 
              Style.RESET_ALL)


	if test_number == 3:

		print("=================\n*** QUICK TEST 1 ***\n=================\n")

		f = TEST_RDF_FOLDER + "pizza.ttl"

		o = Ontospy(f, verbose=True, rdf_format="ttl", hide_implicit_types=False, hide_base_schemas=True, hide_implicit_preds=False, hide_individuals=False)
		print(f)
		for x in o.all_individuals:
			print(x.qname, [z.uri for z in x.instance_of()])			





if __name__ == '__main__':
	main()




# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
#
#
# ==================
# VIZ SIGMA JS
# ==================


import os, sys
import json

from ..utils import *
from ..actions import * 
from ..viz_factory import VizFactory





class Dataviz(VizFactory):
    """
    Sigma JS

    """


    def __init__(self, ontospy_graph, title=""):
        """
        Init
        """
        super(Dataviz, self).__init__(ontospy_graph, title)
        self.static_files = ["libs/sigma", "libs/jquery"]


    def _buildTemplates(self):
        """
        OVERRIDING THIS METHOD from Factory
        """

        c_mydict = build_class_json(self.ontospy_graph.all_classes)
        JSON_DATA_CLASSES = json.dumps(c_mydict)
        c_total = len(self.ontospy_graph.all_classes)


        extra_context = {
                        "ontograph": self.ontospy_graph,
                        'JSON_DATA_CLASSES' : JSON_DATA_CLASSES,
                        "TOTAL_CLASSES": c_total,
                        }

        # Ontology - MAIN PAGE
        contents = self._renderTemplate("sigma/sigmajs.html", extraContext=extra_context)
        FILE_NAME = "index.html"
        main_url = self._save2File(contents, FILE_NAME, self.output_path)

        return main_url




import random

def _classColor(x):
    if not x.parents():
        return "#FFD8DD"
    elif not x.children():
        return "#C4F5FF"
    else:
        return "#C8FFB3"

def build_class_json(classes):
    nodes = [{
                    "id": "owlthing",
                    "label": "owl:Thing",
                    "x": random.random(),
                    "y": random.random(),
                    "color" : "grey",
                    "size": 5
                }]
    edges = []
    for x in classes:
        temp = {
                    "id": x.id,
                    "label": x.bestLabel(quotes=False),
                    "x": random.random(),
                    "y": random.random(),
                    "color" : _classColor(x),
                    "size": len(x.descendants()) or 1
                }
        nodes.append(temp)

        if not x.parents():
            # add a relationship to OWL THING
            edges.append({
                "id": "%s-%s" % (x.id, "owlthing"),
                "source": "owlthing",
                "target": x.id,
                "color": '#ccc',
                "hover_color": '#000'
            })

        for el in x.children():
            temp2 = {
                "id": "%s-%s" % (x.id, el.id),
                "source": x.id,
                "target": el.id,
                 # "type": 'curve',
                "color": '#ccc',
                "hover_color": '#000'
            }
            edges.append(temp2)

    result = {
            "nodes": nodes ,
            "edges" : edges,
        }

    return result





# if called directly, for testing purposes pick a random ontology

if __name__ == '__main__':

    TEST_ONLINE = False
    try:

        g = get_onto_for_testing(TEST_ONLINE)

        v = Dataviz(g, title="")
        v.build()
        v.preview()

        sys.exit(0)

    except KeyboardInterrupt as e:  # Ctrl-C
        raise e





# !/usr/bin/env python
#  -*- coding: UTF-8 -*-
"""
Unit test stub for ontosPy

Run like this:

$ python -m ontospy.tests.test_load_local

"""

from __future__ import print_function

import unittest, os
import time

from .. import *
from .. core import *
from .. core.utils import *

from .context import TEST_RDF_FOLDER


# sanity check
printDebug(f"-------------------\nOntospy {VERSION}\n-------------------")


class TestLoadOntologies(unittest.TestCase):


	def test1_load_locally(self):
		"""
		Check if the ontologies in /RDF folder load ok
		"""

		printDebug(f"""\n=================\n
		\nTEST 1: Loading all ontologies in => {TEST_RDF_FOLDER} 
		\nFor each model detailed entities descriptions are printed out.
		\n\n=================""", bg="blue", fg="white")

		time.sleep(3)

		for f in os.listdir(TEST_RDF_FOLDER):
			if not f.startswith('.'):
				printDebug("\n*****\nTest: loading file... > %s\n*****" % str(f), bg="green")

				o = Ontospy(TEST_RDF_FOLDER + f, verbose=True)

				print("CLASS TREE")
				o.printClassTree()
				print("----------")

				# for c in o.all_classes:
				# 	c.describe()

				# for p in o.all_properties:
				# 	p.describe()

				# for s in o.all_skos_concepts:
				# 	s.describe()


				print("Success.\n")



if __name__ == "__main__":
	unittest.main()



# NOTICE
# This software was produced for the U.S. Government under contract FA8702-22-C-0001,
# and is subject to the Rights in Data-General Clause 52.227-14, Alt. IV (DEC 2007)
# ©2021 The MITRE Corporation. All Rights Reserved.

'''
This module implements a function that determines SHACL property constraints
from the triple provided by Ontospy.

In this file,
    OntoProperties is the ontospy class ontospy.core.entities.OntoProperties
    OntoClass is the ontospy class ontospy.core.entities.OntoClass
'''

from typing import Any, DefaultDict, Dict, List, Optional, Set, Tuple, Union

from collections import defaultdict
import rdflib
from rdflib import BNode, Graph, Literal, RDFS, SH, URIRef
from rdflib.term import Node
    # SH   = http://www.w3.org/ns/shacl#
    # OWL  = http://www.w3.org/2002/07/owl#
    # RDF  = http://www.w3.org/1999/02/22-rdf-syntax-ns#
    # RDFS = http://www.w3.org/2000/01/rdf-schema#


from .entities import OntoClass, OntoProperty, OntoShape, Ontology, RdfEntity

# TODO - This type is defined in what is currently a prerelease state of RDFLib.  Import rdflib.IdentifiedNode once RDFLib's version is >6.1.1.
IdentifiedNode = Union[BNode, URIRef]

# all_classes is a module-level tracking of all of the SHACL classes known to the caller of build_shacl_constraints.
# As a reminder, SHACL classes are explicitly noted as classes (e.g. RDF or OWL Class instances), or are inferred to be classes by appearing as a rdf:type Object.
# https://www.w3.org/TR/shacl/#dfn-shacl-class
# all_classes is indexed by the IRI of the class.  Note that this constrains OntoSpy to only representing classes that are not BNodes, which might or might not be an issue with anonymous classes defined as part of OWL Restrictions.
all_classes: Dict[URIRef, OntoClass] = dict()

# all_properties follows the same practice as all_classes.
all_properties: Dict[URIRef, OntoProperty] = dict()

namespace_manager = rdflib.namespace.NamespaceManager(Graph())


def set_namespace_manager(namespaces) -> None:
    for (prefix, uri) in namespaces:
        namespace_manager.bind(prefix, uri)


class NodeShape(object):
    '''
    Container for a class uri and its corresponding OntoClass object.
    A NodeShape MUST have a uri and may or may not have an OntoClass
    Two NodeShape objects are EQUAL if they have the same self.class_uri

    Attributes:
        self.class_uri       The URI of an ontology class
        self.onto_class      The OntoClass object with uri self.class_uri,
                                or None if this class is not in the ontology
        self.qname           The qname (onto_class.qname if onto_class is not None)
    '''
    # Class constant

    def __init__(self, class_uri: rdflib.URIRef) -> None:
        self.class_uri = class_uri
        self.onto_class = all_classes.get(class_uri)  # May be None
        self.qname = namespace_manager.qname(class_uri)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.class_uri == other.class_uri
        else:
            return False

    def __hash__(self):
        return hash(self.class_uri)

    def __str__(self):
        return '<{}, {}>'.format(self.qname, self.onto_class)
    def __repr__(self):
        return str(self)



class Property(object):
    '''
    Container for a property uri and its corresponding OntoProperty object.
    A Property MUST have a uri and may or may not have an OntoProperty
    Two Property objects are EQUAL if they have the same self.property_uri

    Attributes:
        self.property_uri       The URI of an ontology property
        self.onto_property      The OntoProperty object with uri self.property_uri,
                                or None if this property is not in the ontology
    '''

    def __init__(self, property_uri: rdflib.URIRef) -> None:
        self.property_uri = property_uri
        self.onto_property: Optional[OntoProperty] = all_properties.get(property_uri)
        self.qname: str
        try:
            self.qname = namespace_manager.qname(property_uri)
        except:
            self.qname = str(property_uri)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.property_uri == other.property_uri
        else:
            return False

    def __hash__(self):
        return hash(self.property_uri)

    def __str__(self):
        return '<{}, {}>'.format(self.property_uri, self.onto_property)
    def __repr__(self):
        return str(self)



class Constraint(object):
    '''
    Container for constraint attributes.

    Attributes are SETS of values.  Although there should normally be only zero
    or one value for each attribute, we need to be able to handle the
    possibility of there being multiple values.

    A Header Constraint is a Constraint object that contains only a header string
    If the header argument is specified, the property_obj is ignored and a Header Constraint is returned
    '''
    def __init__(self, property_obj=None, header=None) -> None:
        '''
        Create empty instance of this class.

        property_obj    The onto_prop object for this Constraint object to implement
        header          If present, this is a Header Constraint.  Ignore the property_obj.
        '''
        if header:
            self.header = header
            return

        self.property_obj = property_obj   # PROPERTY: Property object
        self.sh_path = property_obj.qname  # THIS PROPERTY, a Property object
        self.sh_minCount: Set[str] = set()
        self.sh_maxCount: Set[str] = set()
        self.sh_datatype: Set[NodeShape] = set()
        self.sh_class: Set[NodeShape] = set()
        self.sh_description: Set[str] = set()
        self.sh_minInclusive: Set[str] = set()
        self.sh_maxInclusive: Set[str] = set()
        self.sh_minExclusive: Set[str] = set()
        self.sh_maxExclusive: Set[str] = set()
        self.sh_minLength: Set[str] = set()
        self.sh_maxLength: Set[str] = set()
        self.sh_pattern: Set[str] = set()
        self.sh_disjoint: Set[RdfEntity] = set()            #  Property objects
        self.sh_equals: Set[OntoProperty] = set()              #  Property objects
        self.sh_lessThan: Set[OntoProperty] = set()            #  Property objects
        self.sh_lessThanOrEquals: Set[OntoProperty] = set()    #  Property objects
        self.sh_not: Set[NodeShape] = set()
        self.sh_hasValue: Set[Union[str, RdfEntity]] = set()
        self.sh_in: Set[RdfEntity] = set()
        self.rdftype_qname: Set[str] = set()       # "owl:ObjectProperty" exclusive-or "owl:DatatypeProperty"
        self.rdfs_range: Set[NodeShape] = set()
        self.rdfs_comment: Set[str] = set()


    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)



def build_shacl_constraints(ontology_object: Ontology) -> Dict[OntoClass, List[Constraint]]:
    '''
    Arguments:
        ontology_object   An Ontospy Ontology object

    Return:  All SHACL Constraint Dictionary
        all_shacl_constraints is {onto_class:[Constraint]}
        and the list of Constraint objects are sorted by property_uri.qname

    Ontodoc's job is to display the ontology, not to validate it.  If there are conflicting
    property ranges or cardinalities, all values are displayed.

    The display has eight columns:  PROPERTY, TYPE, DESCRIPTION, MIN, MAX, OTHER CONSTRAINTS, LOCAL TYPE, GLOBAL RANGE

    (1) Each sh:property entry describes a single property.
        The sh:property
        (a) MUST have a sh:path property, which is a property URI.  This is saved as property_uri and goes in the PROPERTY column.
        (b) MAY have sh:minCount, sh:maxCount.  This is saved as min_count and max_count, and goes into the MIN and MAX columns.
        (c) MAY have sh:class.  This is saved as a NodeShape object in range_obj and goes in the RANGE column, and TYPE is owl:ObjectProperty.
        (d) MAY have sh:datatype.  This is saved as a NodeShape object in range_obj, and goes in the RANGE column, and TYPE is owl:DatatypeProperty.

    (2) Each sh:path property MUST defined as an OWL property declaration in the ontology.
        The OWL property description
        (a) MUST have an rdftype_qname that is either "owl:ObjectProperty" or "owl:DatatypeProperty".  This is saved as property_type and goes in the TYPE column.
        (b) SHOULD have an rdfs:comment property.  This is saved as description and goes in the DESCRIPTION column.
        (c) SHOULD have an rdfs:range property.  This is saved as a NodeShape object in range_obj, and goes the RANGE column.

    Display notes (these defaults are implemented in the template, and may actually be implemented differently)
        If there is no property_uri, no property is displayed.
        If there is no property_type, the displayed value for TYPE is blank (this should never happen)
        If there is no description, the displayed value for DESCRIPTION is blank
        If there is no range_obj, the displayed value for RANGE is ("rdfs resource", None)
        If there is no min_count, the displayed value for MIN is '0'.
        If there is no max_count, the displayed value for MAX is '*'
    '''

    # Start with empty results
    all_shacl_constraints: Dict[OntoClass, List[Constraint]] = dict()

    # Compute {property_uri:OntoProperty}
    for onto_prop in ontology_object.all_properties:
        all_properties[onto_prop.uri] = onto_prop

    set_namespace_manager(ontology_object.namespaces)

    # Populate the dictionary mapping class IRIs to OntoClass objects.
    ontology_class: OntoClass
    for ontology_class in ontology_object.all_classes:
        all_classes[ontology_class.uri] = ontology_class

    # Do for each class in the ontology
    onto_class: OntoClass
    for onto_class in ontology_object.all_classes:

        # Start with an empty LIST of ordered Constraint objects for this onto_class
        constraints: List[Constraint] = []

        # Do for each class in the lineage (self, parents, grandparents, etc)
        lineage_classes: List[OntoClass] = get_lineage(onto_class)
        lineage_class: OntoClass
        for lineage_class in lineage_classes:

            # If class has no shapes, skip
            if not lineage_class.shapedProperties:
                continue

            # Start with empty dictionary of property constraints for this lineage class
            # The key of this dictionary is a property IRI (that is, a rdflib.URIRef).
            property_constraints: Dict[URIRef, Constraint] = dict()

            # Do for each shape in this class (multiple shapes may have constraints for the same property_uri)

            for shaped_property_dict in lineage_class.shapedProperties:
                assert isinstance(shaped_property_dict["shape"], OntoShape)
                shape: OntoShape = shaped_property_dict["shape"]
                # Parse triples to get property constraints for this shape and add to property_constraints
                add_property_constraints_from_shape_triples(lineage_class.uri, property_constraints, shape.triples)

            # Add property constraints from owl property descriptions to constraints
            add_owl_property_constraints(property_constraints)

            # If there's a header, add to LIST of constraints
            #if lineage_class != onto_class:
            #    constraints.append(Constraint(header = lineage_class.qname))

            # Always add header to LIST of constraints
            constraints.append(Constraint(header = lineage_class.qname))

            # Add constraints to LIST of constraints in alphabetical order
            constraints.extend(sorted(property_constraints.values(), key=lambda v:v.property_obj.qname))

        # Add cumulative constraints to all_shacl_constraints
        all_shacl_constraints[onto_class] = constraints

    # Return all the constraints
    return all_shacl_constraints


def add_property_constraints_from_shape_triples(
    class_uri: URIRef,
    property_constraints: Dict[URIRef, Constraint],
    shacl_triples: Optional[List[Tuple[IdentifiedNode, IdentifiedNode, Node]]]
) -> None:
    '''
    Arguments:
        class_uri                 URI of current class
        property_constraints      {property_uri:Constraint}
        shacl_triples             List of shacl triples [(s, p, o)] pertaining to sh:PropertyShapes attached to the current class.

    Action:
        Modify property_constraints by property_uris and their Constraint object derived from the shacl triples
    '''
    # If there are no triples, do nothing
    if not shacl_triples:
        return

    # Build temporary graph to get access to RDFLib iterators.
    tmp_graph = Graph()
    for shacl_triple in shacl_triples:
        tmp_graph.add(shacl_triple)

    # From sh:path, get property URI when object of sh:path triple is a named concept.
    # Once found, property_uri will be used to guarantee an entry for property_constraints.
    for n_property_shape, pred, obj in shacl_triples:
        if pred != SH["path"]:
            continue
        if not isinstance(obj, URIRef):
            # (sh:path can be an RDF List of predicates to follow in-sequence.  Designing documentation for that is left as future work.)
            continue
        property_uri: URIRef = obj
        
        # Find this property's Constraint object in property_constraints.
        # If it hasn't been defined yet, create one.
        if property_uri not in property_constraints:
            property_constraints[property_uri] = Constraint(property_obj=Property(property_uri))
        constraint: Constraint = property_constraints[property_uri]

        # Convert values in graph to sets of values to consolidate and/or compare after looping through lineage (outside scope of this subroutine).

        for triple in tmp_graph.triples((n_property_shape, SH["class"], None)):
            constraint.sh_class.update({NodeShape(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["datatype"], None)):
            constraint.sh_datatype.update({NodeShape(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["description"], None)):
            constraint.sh_description.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["disjoint"], None)):
            constraint.sh_disjoint.update({OntoProperty(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["equals"], None)):
            constraint.sh_equals.update({OntoProperty(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["hasValue"], None)):
            if isinstance(triple[2], Literal):
                constraint.sh_hasValue.update({str(triple[2])})
            else:
                constraint.sh_hasValue.update({RdfEntity(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["in"], None)):
            constraint.sh_in.update({RdfEntity(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["lessThan"], None)):
            constraint.sh_lessThan.update({OntoProperty(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["lessThanOrEquals"], None)):
            constraint.sh_lessThanOrEquals.update({OntoProperty(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["minCount"], None)):
            constraint.sh_minCount.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["maxCount"], None)):
            constraint.sh_maxCount.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["maxExclusive"], None)):
            constraint.sh_maxExclusive.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["maxInclusive"], None)):
            constraint.sh_maxInclusive.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["maxLength"], None)):
            constraint.sh_maxLength.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["minExclusive"], None)):
            constraint.sh_minExclusive.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["minInclusive"], None)):
            constraint.sh_minInclusive.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["minLength"], None)):
            constraint.sh_minLength.update({str(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["not"], None)):
            constraint.sh_not.update({NodeShape(triple[2])})

        for triple in tmp_graph.triples((n_property_shape, SH["pattern"], None)):
            constraint.sh_pattern.update({str(triple[2])})



def add_owl_property_constraints(property_constraints: Dict[URIRef, Constraint]) -> None:
    '''
    Arguments:
        property_constraints    {property_uri:Constraint}

    Action:
        Modify property_constraints by adding additional data
    '''
    # Do for each property_uri in cumulative constraints
    for property_uri, constraint in property_constraints.items():

        # Get OntoProperty object.  If no such object, do nothing
        property_obj = Property(property_uri)
        if property_obj.onto_property is None:
            return
        onto_property = property_obj.onto_property

        # From OntoProperty.rdftype_qname, add property_type
        constraint.rdftype_qname.add(onto_property.rdftype_qname)   # "owl:DatatypeProperty" or "owl:ObjectProperty"

        if onto_property.triples is None:
            return

        # Build temporary graph to get access to RDFLib iterators.
        tmp_graph = Graph()
        for rdfs_triple in onto_property.triples:
            tmp_graph.add(rdfs_triple)

        # From rdfs:comment, add comments from owl constraint to Constraints
        for triple in tmp_graph.triples((property_uri, RDFS["comment"], None)):
            constraint.rdfs_comment.update({str(triple[2])})

        # From rdfs:range, add NodeShape object to Constraints (skip BNodes)
        for triple in tmp_graph.triples((property_uri, RDFS["range"], None)):
            if not isinstance(triple[2], URIRef):
                continue
            constraint.rdfs_range.update({NodeShape(triple[2])})


def get_lineage(onto_class: OntoClass) -> List[OntoClass]:
    '''
    Arguments:
        onto_class    An ontoClass object

    Return:
        LIST containing onto_class, its parents, grandparents, etc, in breadth-first order.
        The first item in the returned list is always the input, onto_class
        Remaining items are ancestor classes in breadth-first search order
    '''
    # Start with self
    onto_classes: List[OntoClass] = [onto_class]

    # Add ancestors
    # Inner class -- recursively add ancestors, one generation at a time (breadth-first)
    def _add_ancestors(sibling_classes: List[OntoClass]) -> None:
        '''
        Arguments:
            sibling_classes  LIST of ontoClass objects whose parents to add

        Closure:
            onto_classes   Accumulating list of ancestor classes in breadth first order
        '''
        # Get next generation
        parent_classes: List[OntoClass] = []
        for sibling_class in sibling_classes:
            for parent_class in sibling_class.parents():
                assert isinstance(parent_class, OntoClass)
                parent_classes.append(parent_class)

        # If no parents, done
        if not parent_classes:
            return

        # Add parents in alphabetical order
        parent_classes.sort(key=lambda x:x.qname)
        onto_classes.extend(parent_classes)

        # Call self with parents
        _add_ancestors(parent_classes)


    _add_ancestors(onto_classes)

    # Remove duplicates while maintaining order
    unique_onto_classes: List[OntoClass] = []
    for onto_class in onto_classes:
        if onto_class not in unique_onto_classes:
            unique_onto_classes.append(onto_class)

    # Return results
    return unique_onto_classes



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-


"""
ONTOSPY
Copyright (c)  __Michele Pasin__ <http://www.michelepasin.org>. All rights reserved.
"""

from itertools import chain
import rdflib

from .utils import *
from .rdf_loader import RDFLoader
from .entities import *
from .sparql_helper import SparqlHelper
from .shacl_helper import build_shacl_constraints


class Ontospy(object):
	"""
	Object that extracts schema definitions (aka 'ontologies') from an rdf graph.

	In [3]: import ontospy

	In [5]: o = ontospy.Ontospy()

	In [7]: o.load_rdf("foaf.rdf")

	In [11]: o.build_all()

	In [13]: o.stats()
	Out[13]:
	[('Ontologies', 1),
	 ('Triples', 630),
	 ('Classes', 15),
	 ('Properties', 67),
	 ('Annotation Properties', 7),
	 ('Object Properties', 34),
	 ('Datatype Properties', 26),
	 ('Skos Concepts', 0),
	 ('Data Sources', 1)]

	"""

	def __init__(self,
				 uri_or_path=None,
				 data=None,
				 file_obj=None,
				 rdf_format="",
				 verbose=False,
				 hide_base_schemas=True,
				 hide_implicit_types=True,
				 hide_implicit_preds=True,
				 hide_individuals=True,
				 sparql_endpoint=None,
				 credentials=None,
				 build_all=True,
				 pref_title="qname",
				 pref_lang="en",
				 ):
		"""Load the graph in memory, then setup all necessary attributes.

		Parameters
		----------
		uri_or_path : [type], optional
			[description], by default None
		data : [type], optional
			[description], by default None
		file_obj : [type], optional
			[description], by default None
		rdf_format : str, optional
			[description], by default ""
		verbose : bool, optional
			[description], by default False
		hide_base_schemas : bool, optional
			[description], by default True
		hide_implicit_types : bool, optional
			[description], by default True
		hide_implicit_preds : bool, optional
			[description], by default True
		hide_individuals: bool, opt
			Extract class instances or not, by default True
		sparql_endpoint : [type], optional
			[description], by default None
		credentials : [type], optional
			[description], by default None
		build_all : bool, optional
			[description], by default True
		pref_title : str, optional
			How to display entities by default. Two options accepted: "qname" (default) or "label" for rdfs:label. 
		pref_lang : str, optional
			Default: 'en'
		"""
		
		super(Ontospy, self).__init__()

		self.rdflib_graph = None
		self.sparql_endpoint = None
		self.credentials = None  # tuple: auth credentials for endpoint if needed
		self.sources = None
		self.sparqlHelper = None
		self.pref_title = pref_title
		self.pref_lang = pref_lang
		self.hide_individuals = hide_individuals
		self.namespaces = []
		# entities buckets start with 'all_'
		self.all_ontologies = []
		self.all_classes = []
		self.all_properties = []
		self.all_properties_annotation = []
		self.all_properties_object = []
		self.all_properties_datatype = []
		self.all_skos_concepts = []
		self.all_shapes = []
		self.all_individuals = []
		self.toplayer_classes = []
		self.toplayer_properties = []
		self.toplayer_skos = []
		self.toplayer_shapes = []
		self.OWLTHING = OntoClass(rdflib.OWL.Thing, 
									rdflib.OWL.Class, 
									self.namespaces,
									False, 
									self.pref_title, 
									self.pref_lang)

		# finally:
		if uri_or_path or data or file_obj:
			self.load_rdf(uri_or_path, data, file_obj, rdf_format, verbose)
			if build_all:
				self.build_all(
					verbose=verbose,
					hide_base_schemas=hide_base_schemas,
					hide_implicit_types=hide_implicit_types,
					hide_implicit_preds=hide_implicit_preds,
					hide_individuals=hide_individuals,
					)
		elif sparql_endpoint:  # by default entities are not extracted
			self.load_sparql(sparql_endpoint, verbose, credentials)
		else:
			pass

	def __repr__(self):
		"""
		Return some info for the ontospy instance.

		note: if it's a sparql backend, limit the info returned to avoid long queries (tip: a statement like `if self.rdflib_graph` on a sparql endpoint is enough to cause a long query!)

		"""
		if self.sparql_endpoint and self.rdflib_graph != None:
			return "<Ontospy Graph (sparql endpoint = <%s>)>" % self.sparql_endpoint
		elif self.rdflib_graph != None:
			return "<Ontospy Graph (%d triples)>" % (len(self.rdflib_graph))
		else:
			return "<Ontospy object created but not initialized (use the `load_rdf` method to load an rdf schema)>"

	def load_rdf(self,
				 uri_or_path=None,
				 data=None,
				 file_obj=None,
				 rdf_format="",
				 verbose=False):
		"""Load an RDF source into an ontospy/rdflib graph"""
		loader = RDFLoader(verbose=verbose)
		loader.load(uri_or_path, data, file_obj, rdf_format)
		self.rdflib_graph = loader.rdflib_graph
		self.sources = loader.sources_valid
		self.sparqlHelper = SparqlHelper(self.rdflib_graph)
		self.namespaces = sorted(self.rdflib_graph.namespaces())

	def load_sparql(self,
					sparql_endpoint,
					verbose=False,
					credentials=None):
		"""
		Set up a SPARQLStore backend as a virtual ontospy graph

		Note: we're using a 'SPARQLUpdateStore' backend instead of 'SPARQLStore' cause otherwise authentication fails (https://github.com/RDFLib/rdflib/issues/755)

		@TODO this error seems to be fixed in upcoming rdflib versions
		https://github.com/RDFLib/rdflib/pull/744

		"""
		try:
			# graph = rdflib.Graph('SPARQLStore')
			# graph = rdflib.ConjunctiveGraph('SPARQLStore')
			graph = rdflib.ConjunctiveGraph('SPARQLUpdateStore')

			if credentials and type(credentials) == tuple:
				# https://github.com/RDFLib/rdflib/issues/343
				graph.store.setCredentials(credentials[0], credentials[1])
				# graph.store.setHTTPAuth('BASIC') # graph.store.setHTTPAuth('DIGEST')

			graph.open(sparql_endpoint)
			self.rdflib_graph = graph
			self.sparql_endpoint = sparql_endpoint
			self.sources = [sparql_endpoint]
			self.sparqlHelper = SparqlHelper(self.rdflib_graph, self.sparql_endpoint)
			self.namespaces = sorted(self.rdflib_graph.namespaces())
		except:
			printDebug("Error trying to connect to Endpoint.")
			raise
		# don't extract entities by default..


	# ------------
	# === methods to build python objects === #
	# ------------

	def build_all(self,
				  verbose=False,
				  hide_base_schemas=True,
				  hide_implicit_types=True,
				  hide_implicit_preds=True,
				  hide_individuals=True,
				  ):
		"""
		Extract all ontology entities from an RDF graph and construct Python representations of them.
		"""
		if verbose:
			printInfo("Scanning entities...", "green")
			printInfo("----------", "comment")

		self.build_ontologies()
		if verbose:
			printInfo("Ontologies.........: %d" % len(self.all_ontologies), "comment")

		self.build_classes(hide_base_schemas, hide_implicit_types)
		if verbose:
			printInfo("Classes............: %d" % len(self.all_classes), "comment")

		self.build_properties(hide_implicit_preds)
		if verbose:
			printInfo("Properties.........: %d" % len(self.all_properties), "comment")
		if verbose:
			printInfo("..annotation.......: %d" % len(self.all_properties_annotation), "comment")
		if verbose:
			printInfo("..datatype.........: %d" % len(self.all_properties_datatype), "comment")
		if verbose:
			printInfo("..object...........: %d" % len(self.all_properties_object), "comment")

		self.build_skos_concepts()
		if verbose:
			printInfo("Concepts (SKOS)....: %d" % len(self.all_skos_concepts), "comment")

		self.build_shapes()
		if verbose:
			printInfo("Shapes (SHACL).....: %d" % len(self.all_shapes), "comment")

		if not hide_individuals:
			self.build_individuals()
			if verbose:
				printInfo("Individuals........: %d" % len(self.all_individuals), "comment")			

		# self.__computeTopLayer()

		self.__computeInferredProperties()

		if verbose:
			printInfo("----------", "comment")

	def build_ontologies(self, exclude_BNodes=False, return_string=False):
		"""
		Extract ontology instances info from the graph, then creates python objects for them.

		Note: often ontology info is nested in structures like this:

		[ a owl:Ontology ;
			vann:preferredNamespacePrefix "bsym" ;
			vann:preferredNamespaceUri "http://bsym.bloomberg.com/sym/" ]

		Hence there is some logic to deal with these edge cases.
		"""
		out = []

		qres = self.sparqlHelper.getOntology()

		if qres:
			# NOTE: SPARQL returns a list of rdflib.query.ResultRow (~ tuples..)

			for candidate in qres:
				if isBlankNode(candidate[0]):
					if exclude_BNodes:
						continue
					else:
						checkDC_ID = [x for x in self.rdflib_graph.objects(
							candidate[0], rdflib.namespace.DC.identifier)]
						if checkDC_ID:
							out += [Ontology(checkDC_ID[0], 
									namespaces=self.namespaces, 
									pref_title=self.pref_title), 
									]
						else:
							vannprop = rdflib.URIRef(
								"http://purl.org/vocab/vann/preferredNamespaceUri")
							vannpref = rdflib.URIRef(
								"http://purl.org/vocab/vann/preferredNamespacePrefix")
							checkDC_ID = [x for x in self.rdflib_graph.objects(
								candidate[0], vannprop)]
							if checkDC_ID:
								checkDC_prefix = [
									x for x in self.rdflib_graph.objects(candidate[0], vannpref)]
								if checkDC_prefix:
									out += [Ontology(checkDC_ID[0],
													 namespaces=self.namespaces,
													 pref_prefix=checkDC_prefix[0],
													 pref_title=self.pref_title,
													 pref_lang=self.pref_lang,
													 )
													 ]
								else:
									out += [Ontology(checkDC_ID[0], 
											namespaces=self.namespaces, 
											pref_title=self.pref_title,
											pref_lang=self.pref_lang,
											)]

				else:
					out += [Ontology(candidate[0], 
								namespaces=self.namespaces, 
								pref_title=self.pref_title,
								pref_lang=self.pref_lang,
								)]

		else:
			pass
			# printInfo("No owl:Ontologies found")

		# finally... add all annotations/triples
		self.all_ontologies = out
		for onto in self.all_ontologies:
			onto.triples = self.sparqlHelper.entityTriples(onto.uri)
			onto._buildGraph()  # force construction of mini graph

		# sort alphabetically
		self.all_ontologies = sorted(self.all_ontologies, key=lambda x: x.uri)


	#
	#  RDFS:class vs OWL:class cf. http://www.w3.org/TR/owl-ref/ section 3.1
	#

	def build_classes(self, hide_base_schemas=True, hide_implicit_types=True):
		"""
		2015-06-04: removed sparql 1.1 queries
		2015-05-25: optimized via sparql queries in order to remove BNodes
		2015-05-09: new attempt

		Note: sparqlHelper.getAllClasses() returns a list of tuples,
		(class, classRDFtype)
		so in some cases there are duplicates if a class is both RDFS.CLass and OWL.Class
		In this case we keep only OWL.Class as it is more informative.
		"""

		self.all_classes = []  # @todo: keep adding?

		qres = self.sparqlHelper.getAllClasses(hide_base_schemas,
											   hide_implicit_types)

		for class_tuple in qres:

			_uri = class_tuple[0]
			try:
				_type = class_tuple[1]
			except:
				_type = ""

			test_existing_cl = self.get_class(uri=_uri)
			if not test_existing_cl:
				# create it
				ontoclass = OntoClass(_uri, _type, 
									 self.namespaces, 
									 False,
									 self.pref_title, 
									 self.pref_lang)
				self.all_classes += [ontoclass]
			else:
				# if OWL.Class over RDFS.Class - update it
				if _type == rdflib.OWL.Class:
					test_existing_cl.rdftype = rdflib.OWL.Class


		# add more data
		for aClass in self.all_classes:
			# print("enriching class", aClass)
			aClass.triples = self.sparqlHelper.entityTriples(aClass.uri)
			aClass._buildGraph()  # force construction of mini graph

			aClass.sparqlHelper = self.sparqlHelper

			# attach to an ontology
			for uri in aClass.getValuesForProperty(rdflib.RDFS.isDefinedBy):
				onto = self.get_ontology(uri=str(uri))
				if onto:
					onto.all_classes += [aClass]
					aClass.ontology = onto

			# add direct Supers
			directSupers = self.sparqlHelper.getClassDirectSupers(aClass.uri)

			for x in directSupers:
				superclass = self.get_class(uri=x[0])
				# note: extra condition to avoid recursive structures
				if superclass and superclass.uri != aClass.uri:
					aClass._parents.append(superclass)

					# add inverse relationships (= direct subs for superclass)
					if aClass not in superclass.children():
						superclass._children.append(aClass)

		# sort alphabetically
		self.all_classes = sorted(self.all_classes, key=lambda x: x.qname)

		# compute top layer
		# print("calc toplayer")
		exit = []
		for c in self.all_classes:
			if not c.parents():
				exit += [c]
		self.toplayer_classes = exit  # sorted(exit, key=lambda x: x.id) # doesnt work

	def build_properties(self, hide_implicit_preds=True):
		"""
		2015-06-04: removed sparql 1.1 queries
		2015-06-03: analogous to get classes

		# instantiate properties making sure duplicates are pruned
		# but the most specific rdftype is kept
		# eg OWL:ObjectProperty over RDF:property

		"""
		self.all_properties = []  # @todo: keep adding?
		self.all_properties_annotation = []
		self.all_properties_object = []
		self.all_properties_datatype = []

		qres = self.sparqlHelper.getAllProperties(hide_implicit_preds)
		# print("rdflib query done")

		for candidate in qres:

			test_existing_prop = self.get_property(uri=candidate[0])
			if not test_existing_prop:
				# create it
				self.all_properties += [OntoProperty(candidate[0], candidate[1], 
											self.namespaces, 
											False,
											self.pref_title,
											self.pref_lang,
											)]
			else:
				# update it
				if candidate[1] and (test_existing_prop.rdftype == rdflib.RDF.Property):
					test_existing_prop.rdftype = inferMainPropertyType(candidate[1])
		# print("properties created")

		# add more data
		for aProp in self.all_properties:
			# print("enriching prop..", aProp)
			if aProp.rdftype == rdflib.OWL.DatatypeProperty:
				self.all_properties_datatype += [aProp]
			elif aProp.rdftype == rdflib.OWL.AnnotationProperty:
				self.all_properties_annotation += [aProp]
			elif aProp.rdftype == rdflib.OWL.ObjectProperty:
				self.all_properties_object += [aProp]
			else:
				pass

			aProp.triples = self.sparqlHelper.entityTriples(aProp.uri)
			aProp._buildGraph()  # force construction of mini graph

			# attach to an ontology [2015-06-15: no property type distinction yet]
			for uri in aProp.getValuesForProperty(rdflib.RDFS.isDefinedBy):
				onto = self.get_ontology(uri=str(uri))
				if onto:
					onto.all_properties += [aProp]
					aProp.ontology = onto

			self.__buildDomainRanges(aProp)

			# add direct Supers
			directSupers = self.sparqlHelper.getPropDirectSupers(aProp.uri)

			for x in directSupers:
				superprop = self.get_property(uri=x[0])
				# note: extra condition to avoid recursive structures
				if superprop and superprop.uri != aProp.uri:
					aProp._parents.append(superprop)

					# add inverse relationships (= direct subs for superprop)
					if aProp not in superprop.children():
						superprop._children.append(aProp)

		# sort alphabetically
		self.all_properties = sorted(self.all_properties, key=lambda x: x.qname)

		# computer top layer for properties
		# print("calc toplayer")
		exit = []
		for c in self.all_properties:
			if not c.parents():
				exit += [c]
		self.toplayer_properties = exit  # sorted(exit, key=lambda x: x.id) # doesnt work

	def build_skos_concepts(self):
		"""
		2015-08-19: first draft
		"""
		self.all_skos_concepts = []  # @todo: keep adding?

		qres = self.sparqlHelper.getSKOSInstances()
		# print("rdflib query done")

		for candidate in qres:

			test_existing_cl = self.get_skos(uri=candidate[0])
			if not test_existing_cl:
				# create it
				self.all_skos_concepts += [OntoSKOSConcept(candidate[0], None, 
											self.namespaces, 
											None, 
											self.pref_title, 
											self.pref_lang,)
											]
			else:
				pass
		# print("concepts created")
		# add more data
		skos = rdflib.Namespace('http://www.w3.org/2004/02/skos/core#')

		for aConcept in self.all_skos_concepts:
			# print("enriching concept...", aConcept)
			aConcept.rdftype = skos['Concept']
			aConcept.triples = self.sparqlHelper.entityTriples(aConcept.uri)
			aConcept._buildGraph()  # force construction of mini graph

			aConcept.sparqlHelper = self.sparqlHelper

			# attach to an ontology
			for uri in aConcept.getValuesForProperty(rdflib.RDFS.isDefinedBy):
				onto = self.get_ontology(uri=str(uri))
				if onto:
					onto.all_skos_concepts += [aConcept]
					aConcept.ontology = onto

			# add direct Supers
			directSupers = self.sparqlHelper.getSKOSDirectSupers(aConcept.uri)

			for x in directSupers:
				superclass = self.get_skos(uri=x[0])
				# note: extra condition to avoid recursive structures
				if superclass and superclass.uri != aConcept.uri:
					aConcept._parents.append(superclass)

					# add inverse relationships (= direct subs for superclass)
					if aConcept not in superclass.children():
						superclass._children.append(aConcept)

		# sort alphabetically
		self.all_skos_concepts = sorted(self.all_skos_concepts, key=lambda x: x.qname)

		# compute top layer for skos
		exit = []
		for c in self.all_skos_concepts:
			if not c.parents():
				exit += [c]
		self.toplayer_skos = exit  # sorted(exit, key=lambda x: x.id) # doesnt work

	def build_shapes(self):
		"""
		Extract SHACL data shapes from the rdf graph.
		<http://www.w3.org/ns/shacl#>

		Instatiate the Shape Python objects and relate it to existing classes,
		if available.
		"""
		self.all_shapes = [] 

		qres = self.sparqlHelper.getShapes()

		for candidate in qres:

			test_existing_cl = self.get_shapes(uri=candidate[0])
			if not test_existing_cl:
				# create it
				self.all_shapes += [OntoShape(candidate[0], None, 
									self.namespaces, 
									None, 
									self.pref_title, 
									self.pref_lang,)
									]
			else:
				pass

		# add more data
		shacl = rdflib.Namespace('http://www.w3.org/ns/shacl#')

		for aShape in self.all_shapes:

			aShape.rdftype = shacl['Shape']
			aShape.triples = self.sparqlHelper.entityTriples(aShape.uri)
			aShape._buildGraph()  # force construction of mini graph

			aShape.sparqlHelper = self.sparqlHelper

			# attach to a class
			for uri in aShape.getValuesForProperty(shacl['targetClass']):
				aclass = self.get_class(str(uri))
				if aclass:
					aShape.targetClasses += [aclass]
					aclass.all_shapes += [aShape]

					#add shaped properties of this class. later can be used for ontodocs
					for propertyUri in aShape.getValuesForProperty(shacl['path']): 
						propType = self.get_property(str(propertyUri))
						if propType:
							aclass.shapedProperties += [{'shape': aShape, 'property': propType}]



		# sort alphabetically
		self.all_shapes = sorted(self.all_shapes, key=lambda x: x.qname)

		# compute top layer
		exit = []
		for c in self.all_shapes:
			if not c.parents():
				exit += [c]
		self.toplayer_shapes = exit 

		# compute SHACL property constraints once classes and shapes have been built
		shacl_constraints = build_shacl_constraints(self)
		for onto_class in self.all_classes:
			onto_class.shacl_constraints = shacl_constraints.get(onto_class, [])


	def build_individuals(self):
		"""
		By default, only add individuals of classes that have been previously extracted.
		"""
		self.all_individuals = []  

		for c in self.all_classes:
			c._instances = []
			qres = self.sparqlHelper.getClassInstances(c.uri)
            
			for uri in [x[0] for x in qres]:

				test_existing_indiv = self.get_individual(uri=uri)

				if not test_existing_indiv:
					instance = RdfEntity(uri, c.uri, c.namespaces)
					instance.triples = self.sparqlHelper.entityTriples(
						instance.uri)
					instance._buildGraph()  # force construction of mini graph
				else:
					instance = test_existing_indiv
				
				c._instances += [instance]
				instance._instance_of += [c]

				if not test_existing_indiv:
					self.all_individuals += [instance]

		# sort alphabetically
		self.all_individuals = sorted(self.all_individuals, key=lambda x: x.qname)


	def build_entity_from_uri(self, uri, ontospyClass=None):
		"""
		Extract RDF statements having a URI as subject, then instantiate the RdfEntity Python object so that it can be queried further.

		Passing <ontospyClass> allows to instantiate a user-defined RdfEntity subclass.

		NOTE: the entity is not attached to any index. In future version we may create an index for these (individuals?) keeping into account that any existing model entity could be (re)created this way.
		"""
		if not ontospyClass:
			ontospyClass = RdfEntity
		elif not issubclass(ontospyClass, RdfEntity):
			printDebug("Error: <%s> is not a subclass of ontospy.RdfEntity" % str(ontospyClass))
			return None
		else:
			pass
		qres = self.sparqlHelper.entityTriples(uri)
		if qres:
			entity = ontospyClass(rdflib.URIRef(uri), 
							None, 
							self.namespaces, 
							None, 
							self.pref_title, 
							self.pref_lang,)
			entity.triples = qres
			entity._buildGraph()  # force construction of mini graph
			# try to add class info
			test = entity.getValuesForProperty(rdflib.RDF.type)
			if test:
				entity.rdftype = test
				entity.rdftype_qname = [entity._build_qname(x) for x in test]
			return entity
		else:
			return None

	# ------------
	# === methods to refine the ontology structure  === #
	# ------------

	def __buildDomainRanges(self, aProp):
		"""
		extract domain/range details and add to Python objects
		"""

		domains = chain(aProp.rdflib_graph.objects(
			None, rdflib.term.URIRef(u'http://schema.org/domainIncludes')), aProp.rdflib_graph.objects(
			None, rdflib.RDFS.domain))

		ranges = chain(aProp.rdflib_graph.objects(
			None, rdflib.term.URIRef(u'http://schema.org/rangeIncludes')), aProp.rdflib_graph.objects(
			None, rdflib.RDFS.range))

		for x in domains:
			if isBlankNode(x):
				aProp.domains += [RdfEntity(x, None, self.namespaces, is_Bnode=True)]
			else:
				aClass = self.get_class(uri=str(x))
				if aClass:
					aProp.domains += [aClass]
					aClass.domain_of += [aProp]
				else:
					# edge case: it's not an OntoClass instance
					aProp.domains += [OntoClass(x, None, 
										self.namespaces, 
										True,  # ext_model arg
										self.pref_title, 
										self.pref_lang,)]

		for x in ranges:
			if isBlankNode(x):
				aProp.domains += [RdfEntity(x, None, self.namespaces, is_Bnode=True)]
			else:
				aClass = self.get_class(uri=str(x))
				if aClass:
					aProp.ranges += [aClass]
					aClass.range_of += [aProp]
				else:
					# eg a DataType property has xsd:STRING
					# here we're storing an ontospy entities but not adding it to
					# the main index
					aProp.ranges += [OntoClass(x, None, 
									self.namespaces, 
									True, 
									self.pref_title,
									self.pref_lang,)]

	def __computeInferredProperties(self):
		"""

		:return: attach a list of dicts to each class, detailing valid props up the subsumption tree
		"""
		exit = []
		for c in self.all_classes:
			c.domain_of_inferred = self.getInferredPropertiesForClass(c, "domain_of")
			c.range_of_inferred = self.getInferredPropertiesForClass(c, "range_of")

	def getInferredPropertiesForClass(self, aClass, rel="domain_of"):
		"""
		returns all properties valid for a class (as they have it in their domain)
		recursively ie traveling up the descendants tree
		Note: results in a list of dicts including itself
		Note [2]: all properties with no domain info are added at the top as [None, props]

		:return:
		[{<Class *http://xmlns.com/foaf/0.1/Person*>:
			[<Property *http://xmlns.com/foaf/0.1/currentProject*>,<Property *http://xmlns.com/foaf/0.1/familyName*>,
				etc....]},
		{<Class *http://www.w3.org/2003/01/geo/wgs84_pos#SpatialThing*>:
			[<Property *http://xmlns.com/foaf/0.1/based_near*>, etc...]},
			]
		"""
		_list = []

		if rel == "domain_of":
			_list.append({aClass: aClass.domain_of})
			for x in aClass.ancestors():
				if x.domain_of:
					_list.append({x: x.domain_of})

			# add properties from Owl:Thing ie the inference layer

			properties_applicable_by_shapes = {x[0] for x in self.sparqlHelper.getPropsApplicableByShapes()}
			topLevelProps = [p for p in self.all_properties if p.domains == [] and not p.uri in properties_applicable_by_shapes]
			if topLevelProps:
				_list.append({self.OWLTHING: topLevelProps})

		elif rel == "range_of":
			_list.append({aClass: aClass.range_of})
			for x in aClass.ancestors():
				if x.domain_of:
					_list.append({x: x.range_of})

			# add properties from Owl:Thing ie the inference layer

			topLevelProps = [p for p in self.all_properties if p.ranges == []]
			if topLevelProps:
				_list.append({self.OWLTHING: topLevelProps})

		return _list



	# ===============
	# methods for retrieving objects
	# ================

	def _get_defined_entity(self, id=None, uri=None, match=None, candidates_list=None):
		"""Inner method to help matching uris or strings to entities that have been added 
		to Ontospy. 

		candidates_list
			A list of entities, eg self.all_classes, self.all_properties, self.all_skos_concepts
		"""

		if not id and not uri and not match:
			return None

		if type(id) == type("string"):
			uri = id
			id = None
			if not is_http(uri):
				match = uri
				uri = None
		if match:
			if type(match) != type("string"):
				return []
			res = []
			if ":" in match:  # qname
				for x in candidates_list:
					if match.lower() in x.qname.lower():
						res += [x]
			else:
				for x in candidates_list:
					if match.lower() in x.uri.lower():
						res += [x]
			return res
		else:
			for x in candidates_list:
				if id and x.id == id:
					return x
				if uri and x.uri.lower() == uri.lower():
					return x
			return None

	def get_class(self, id=None, uri=None, match=None):
		"""
		Get the saved-class with given ID or via other methods...

		Note: it tries to guess what is being passed..

		In [1]: g.get_class(uri='http://www.w3.org/2000/01/rdf-schema#Resource')
		Out[1]: <Class *http://www.w3.org/2000/01/rdf-schema#Resource*>

		In [2]: g.get_class(10)
		Out[2]: <Class *http://purl.org/ontology/bibo/AcademicArticle*>

		In [3]: g.get_class(match="person")
		Out[3]:
		[<Class *http://purl.org/ontology/bibo/PersonalCommunicationDocument*>,
		 <Class *http://purl.org/ontology/bibo/PersonalCommunication*>,
		 <Class *http://xmlns.com/foaf/0.1/Person*>]

		"""

		return self._get_defined_entity(
			id=id, 
			uri=uri, 
			match=match, 
			candidates_list=self.all_classes)


	def get_property(self, id=None, uri=None, match=None):
		"""Get the saved-property with given ID or via other methods...
		"""
		return self._get_defined_entity(
			id=id, 
			uri=uri, 
			match=match, 
			candidates_list=self.all_properties)


	def get_individual(self, id=None, uri=None, match=None):
		"""Get the saved-individual with given ID or via other methods...
		"""
		return self._get_defined_entity(
			id=id, 
			uri=uri, 
			match=match, 
			candidates_list=self.all_individuals)


	def get_skos(self, id=None, uri=None, match=None):
		"""Get the saved skos concept with given ID or via other methods...
		"""
		return self._get_defined_entity(
			id=id, 
			uri=uri, 
			match=match, 
			candidates_list=self.all_skos_concepts)

	def get_shapes(self, id=None, uri=None, match=None):
		"""Get the saved shacl shapes with given ID or via other methods...
		"""
		return self._get_defined_entity(
			id=id, 
			uri=uri, 
			match=match, 
			candidates_list=self.all_shapes)

	def get_ontology(self, id=None, uri=None, match=None):
		"""Get the saved-ontology with given ID or via other methods...
		"""
		return self._get_defined_entity(
			id=id, 
			uri=uri, 
			match=match, 
			candidates_list=self.all_ontologies)

	def get_any_entity(self, id=None, uri=None, match=None):
		"""Get any entity-type with given ID or via other methods...
		"""

		all_entities = self.all_classes + self.all_properties + self.all_individuals + self.all_skos_concepts + self.all_shapes + self.all_ontologies

		return self._get_defined_entity(
			id=id, 
			uri=uri, 
			match=match, 
			candidates_list=all_entities)



	# ===============
	# methods for the CLI interface, so that the user can interact with the ontospy object
	# ================

	def _next_entity(self, uri, candidates_list):
		"""Returns the next entity in the list it is contained in. If it's the last one, returns the first one."""
		if uri == candidates_list[-1].uri:
			return candidates_list[0]
		flag = False
		for x in candidates_list:
			if flag == True:
				return x
			if x.uri == uri:
				flag = True
		return None

	def nextClass(self, classuri):
		"""Returns the next class in the list of classes."""
		return self._next_entity(classuri, self.all_classes)

	def nextProperty(self, propuri):
		"""Returns the next property in the list of classes."""
		return self._next_entity(propuri, self.all_properties)

	def nextConcept(self, concepturi):
		"""Returns the next skos concept in the list of concepts."""
		return self._next_entity(concepturi, self.all_skos_concepts)

	def nextShapes(self, shapeuri):
		"""Returns the next shape in the list of shapes."""
		return self._next_entity(shapeuri, self.all_shapes)

	def nextOntology(self, shapeuri):
		"""Returns the next ontology in the list of ontologies."""
		return self._next_entity(shapeuri, self.all_ontologies)


	# ===============
	# methods for building super/sub trees
	# TODO factor out common patterns into a separate helper method
	# ================


	def _tree_helper(self, entities_list, toplayer):
		"""Common helper method for building super/sub trees"""
		treedict = {}
		if entities_list:
			treedict[0] = toplayer
			for element in entities_list:
				if element.children():
					treedict[element] = element.children()
			return treedict
		return treedict


	def ontologyClassTree(self):
		"""
		Returns a dict representing the ontology class tree
		Top level = {0:[top classes]}
		Multi inheritance is represented explicitly
		"""
		return self._tree_helper(self.all_classes, self.toplayer_classes)

	def ontologyPropTree(self):
		"""
		Returns a dict representing the ontology properties tree
		Top level = {0:[top properties]}
		Multi inheritance is represented explicitly
		"""
		return self._tree_helper(self.all_properties, self.toplayer_properties)

	def ontologyConceptTree(self):
		"""
		Returns a dict representing the skos tree
		Top level = {0:[top concepts]}
		Multi inheritance is represented explicitly
		"""
		return self._tree_helper(self.all_skos_concepts, self.toplayer_skos)

	def ontologyShapeTree(self):
		"""
		Returns a dict representing the shapes tree
		Top level = {0:[top shapes]}
		Multi inheritance is represented explicitly
		"""
		return self._tree_helper(self.all_shapes, self.toplayer_shapes)


	def ontologyIndividualsTree(self):
		"""
		Returns a dict representing the individuals
		Top level = {0:[all individuals]}
		NOTE individuals have no inherent hierarchy, this method was added 
		so to be more consistent with the similar methods for classes etc..
		"""
		treedict = {}
		if self.all_individuals:
			treedict[0] = self.all_individuals
			return treedict
		return treedict


	# ------------
	# === utils === #
	# ------------


	def rdf_source(self, format="turtle"):
		"""
		Wrapper for rdflib serializer method.
		Valid options are: xml, n3, turtle, nt, pretty-xml, json-ld [trix not working out of the box]
		"""
		s = self.rdflib_graph.serialize(format=format)
		if isinstance(s, bytes):
			s = s.decode('utf-8')
		return s

	def serialize(self, format="turtle"):
		"for backward compatibility"
		return self.rdf_source(format)

	def query(self, stringa):
		"""SPARQL query / wrapper for rdflib sparql query method """
		qres = self.rdflib_graph.query(stringa)
		return list(qres)
	def sparql(self, stringa):
		"SPARQL query / replacement for query"
		return self.query(stringa)

	def stats(self):
		""" shotcut to pull out useful info for a graph"""
		out = []
		out += [("Ontologies", len(self.all_ontologies))]
		out += [("Triples", self.triplesCount())]
		out += [("Classes", len(self.all_classes))]
		out += [("Properties", len(self.all_properties))]
		out += [("Annotation Properties", len(self.all_properties_annotation))]
		out += [("Object Properties", len(self.all_properties_object))]
		out += [("Datatype Properties", len(self.all_properties_datatype))]
		out += [("Skos Concepts", len(self.all_skos_concepts))]
		out += [("Data Shapes", len(self.all_shapes))]
		if not self.hide_individuals:
			out += [("Individuals", len(self.all_individuals))]
		out += [("Data Sources", len(self.sources))]
		return out

	def triplesCount(self):
		"""

		2016-08-18 the try/except is a dirty solution to a problem
		emerging with counting graph length on cached Graph objects..
		"""
		# @todo  investigate what's going on..
		# printDebug(unicode(type(self.rdflib_graph)), fg="red")
		try:
			return len(self.rdflib_graph)
		except:
			printDebug("Ontospy: error counting graph length..", fg="red")
			return 0


	def printClassTree(self, element=None, showids=False, labels=False, showtype=False):
		"""
		Print nicely into stdout the class tree of an ontology

		Note: indentation is made so that ids up to 3 digits fit in, plus a space.
		[123]1--
		[1]123--
		[12]12--
		"""
		TYPE_MARGIN = 11  # length for owl:class etc..

		if not element:  # first time
			for x in self.toplayer_classes:
				printGenericTree(x, 0, showids, labels, showtype, TYPE_MARGIN)

		else:
			printGenericTree(element, 0, showids, labels, showtype, TYPE_MARGIN)

	def printPropertyTree(self, element=None, showids=False, labels=False, showtype=False):
		"""
		Print nicely into stdout the property tree of an ontology

		Note: indentation is made so that ids up to 3 digits fit in, plus a space.
		[123]1--
		[1]123--
		[12]12--
		"""
		TYPE_MARGIN = 18  # length for owl:AnnotationProperty etc..

		if not element:  # first time
			for x in self.toplayer_properties:
				printGenericTree(x, 0, showids, labels, showtype, TYPE_MARGIN)

		else:
			printGenericTree(element, 0, showids, labels, showtype, TYPE_MARGIN)

	def printSkosTree(self, element=None, showids=False, labels=False, showtype=False):
		"""
		Print nicely into stdout the SKOS tree of an ontology

		Note: indentation is made so that ids up to 3 digits fit in, plus a space.
		[123]1--
		[1]123--
		[12]12--
		"""
		TYPE_MARGIN = 13  # length for skos:concept

		if not element:  # first time
			for x in self.toplayer_skos:
				printGenericTree(x, 0, showids, labels, showtype, TYPE_MARGIN)

		else:
			printGenericTree(element, 0, showids, labels, showtype, TYPE_MARGIN)

	def printShapesTree(self, element=None, showids=False, labels=False, showtype=False):
		"""
		Print nicely into stdout the SHACL shapes tree of an ontology

		Note: indentation is made so that ids up to 3 digits fit in, plus a space.
		[123]1--
		[1]123--
		[12]12--
		"""
		TYPE_MARGIN = 13  # length for skos:concept

		if not element:  # first time
			for x in self.toplayer_shapes:
				printGenericTree(x, 0, showids, labels, showtype, TYPE_MARGIN)

		else:
			printGenericTree(element, 0, showids, labels, showtype, TYPE_MARGIN)



# !/usr/bin/env python
#  -*- coding: UTF-8 -*-

import os, sys
import json

from ..utils import *
from ..actions import *
from ..viz_factory import VizFactory

# ===========
# D3 PARTITION TABLE
# ===========


class Dataviz(VizFactory):
    """
    D3 Partition table

    """

    def __init__(self, ontospy_graph, title=""):
        """
        Init
        """
        super(Dataviz, self).__init__(ontospy_graph, title)
        self.static_files = ["libs/d3-v3", "libs/jquery"]

    def _buildTemplates(self):
        """
        OVERRIDING THIS METHOD from Factory
        """

        c_total = len(self.ontospy_graph.all_classes)
        p_total = len(self.ontospy_graph.all_properties)
        s_total = len(self.ontospy_graph.all_skos_concepts)

        main_url = None

        if c_total:
            c_mylist = build_D3treeStandard(0, 99, 1,
                                            self.ontospy_graph.toplayer_classes)
            # hack to make sure that we have a default top level object
            JSON_DATA_CLASSES = json.dumps({
                'children': c_mylist,
                'name': 'owl:Thing',
                'id': "None"
            })

            extra_context = {
                "ontograph": self.ontospy_graph,
                "thispage": "classes",
                "TOTAL_CLASSES": c_total,
                "TOTAL_PROPERTIES": p_total,
                "TOTAL_CONCEPTS": s_total,
                "TOTAL_OBJECTS": c_total,
                "TOTAL_OBJECTS_TOPLAYER": len(self.ontospy_graph.toplayer_classes),
                'JSON_DATA_OBJECTS': JSON_DATA_CLASSES,
            }

            # Classes - MAIN PAGE
            contents = self._renderTemplate(
                "d3/d3_partition_table.html", extraContext=extra_context)
            
            if not main_url: 
                FILE_NAME = "index.html"
                main_url = self._save2File(contents, FILE_NAME, self.output_path)
            else:
                FILE_NAME = "classes.html"
                self._save2File(contents, FILE_NAME, self.output_path)
            


        if p_total:
            p_mylist = build_D3treeStandard(0, 99, 1,
                                            self.ontospy_graph.toplayer_properties)
            JSON_DATA_PROPERTIES = json.dumps({
                'children': p_mylist,
                'name': 'Properties',
                'id': "None"
            })

            extra_context = {
                "ontograph": self.ontospy_graph,
                "thispage": "properties",
                "TOTAL_CLASSES": c_total,
                "TOTAL_PROPERTIES": p_total,
                "TOTAL_CONCEPTS": s_total,
                "TOTAL_OBJECTS": p_total,
                "TOTAL_OBJECTS_TOPLAYER": len(self.ontospy_graph.toplayer_properties),
                'JSON_DATA_OBJECTS': JSON_DATA_PROPERTIES,
            }

            # Properties PAGE
            contents = self._renderTemplate(
                "d3/d3_partition_table.html", extraContext=extra_context)
            
            if not main_url: 
                FILE_NAME = "index.html"
                main_url = self._save2File(contents, FILE_NAME, self.output_path)
            else:
                FILE_NAME = "properties.html"
                self._save2File(contents, FILE_NAME, self.output_path)
            

        if s_total:
            s_mylist = build_D3treeStandard(0, 99, 1,
                                            self.ontospy_graph.toplayer_skos)
            JSON_DATA_CONCEPTS = json.dumps({
                'children': s_mylist,
                'name': 'Concepts',
                'id': "None"
            })

            extra_context = {
                "ontograph": self.ontospy_graph,
                "thispage": "concepts",
                "TOTAL_CLASSES": c_total,
                "TOTAL_PROPERTIES": p_total,
                "TOTAL_CONCEPTS": s_total,
                "TOTAL_OBJECTS": s_total,
                "TOTAL_OBJECTS_TOPLAYER": len(self.ontospy_graph.toplayer_skos),
                'JSON_DATA_OBJECTS': JSON_DATA_CONCEPTS,
            }

            # Concepts PAGE
            contents = self._renderTemplate(
                "d3/d3_partition_table.html", extraContext=extra_context)

            if not main_url: 
                FILE_NAME = "index.html"
                main_url = self._save2File(contents, FILE_NAME, self.output_path)
            else:
                FILE_NAME = "skos.html"
                self._save2File(contents, FILE_NAME, self.output_path)


        return main_url

# if called directly, for testing purposes pick a random ontology

if __name__ == '__main__':

    TEST_ONLINE = False
    try:

        g = get_onto_for_testing(TEST_ONLINE)

        v = Dataviz(g, title="")
        v.build()
        v.preview()

        sys.exit(0)

    except KeyboardInterrupt as e:  # Ctrl-C
        raise e
